""" Basic fips sanity test cases for sssd

:requirement: IDM-SSSD-REQ : KRB5 Provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

import time
import re
import pytest
import ldap
from constants import ds_suffix, krb_realm
from sssd.testlib.common.ssh2_python import check_login_client, SSHClient
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.utils import LdapOperations
from sssd.testlib.common.libkrb5 import krb5srv


@pytest.mark.usefixtures('setup_sssd_gssapi', 'create_posix_usersgroups')
@pytest.mark.fips
class Testkrbfips(object):
    """ Testing fips """

    @staticmethod
    @pytest.mark.tier1
    def test_krb_ptr_hash_crash_1792331(multihost):
        """
        :title: sssd_be crashes when krb5_realm and krb5_server
         is omitted and auth_provider is krb5
        :id: b1321b02-4a29-4285-8c85-36f925496463
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1792331
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        multihost.client[0].service_sssd('stop')
        multihost.client[0].run_command(['cat', '/etc/sssd/sssd.conf'])
        kerberos_server = multihost.master[0].sys_hostname
        domain_params = {'krb5_realm': '%s' % krb_realm,
                         'krb5_server': kerberos_server}
        section = 'domain/%s' % domain_name
        tools.sssd_conf(section, domain_params, action='delete')
        start_sssd = 'systemctl start sssd'
        cmd = multihost.client[0].run_command(start_sssd, raiseonerr=False)
        journalctl = 'journalctl -x -n 50 --no-pager'
        crash_msg = "Job for sssd.service failed because a fatal signal"\
                    " was delivered causing the control process to dump core"
        err_msg = "Job for sssd.service failed because the control process "\
                  "exited with error code."
        check1 = re.compile(r'%s' % crash_msg)
        check2 = re.compile(r'%s' % err_msg)
        if check1.search(cmd.stderr_text):
            multihost.client[0].run_command(journalctl)
            tools.sssd_conf(section, domain_params)
            pytest.fail("sssd crashed")
        else:
            tools.sssd_conf(section, domain_params)
            multihost.client[0].run_command(journalctl)
            assert check2.search(cmd.stderr_text)
        multihost.client[0].service_sssd('restart')

    @staticmethod
    def test_fips_login(multihost):
        """
        :title: Verify kerberos user can login successfully in fips mode
        :id: 0ec0efc9-85dd-4a66-9802-65f3b122b7da
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        user = 'foo1@%s' % domain_name
        check_login_client(multihost, user, 'Secret123')

    @staticmethod
    @pytest.mark.tier1_2
    def test_kcm_not_store_tgt(multihost, backupsssdconf):
        """
        :title: sssd-kcm does not store TGT with ssh
         login using GSSAPI
        :id: 9a79474c-26c5-4aeb-9f26-f2ada0e9f453
        :customerscenario: True
        :requirement:
         IDM-SSSD-REQ :: SSSD KCM as default Kerberos CCACHE provider
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1722842
        """
        client = sssdTools(multihost.client[0])
        domain_params = {'debug_level': '10',
                         'ccache_storage': 'memory'}
        client.sssd_conf('kcm', domain_params)
        multihost.client[0].service_sssd('restart')
        multihost.client[0].run_command("systemctl "
                                        "restart sssd-kcm")
        ssh = SSHClient(multihost.client[0].sys_hostname, 'foo3', 'Secret123')
        try:
            ssh.connect()
            ssh.execute_command('kdestroy -A -q')
            ssh.execute_command('echo -e "Secret123" | kinit foo3')
            klist = ssh.execute_command('klist')
            ssh_output = ssh.execute_command(f'ssh -v -o StrictHostKeyChecking=no '
                                             f'-K -l foo3 {multihost.client[0].sys_hostname} klist')
            ssh.close()
        except Exception:
            pytest.fail("Ssh login failed.")

        assert 'KCM:14583103' in klist, "kinit did not work!"
        assert 'KCM:14583103' in ssh_output, "Ticket not forwarded!"

    @staticmethod
    def test_child_logs_after_receiving_hup(multihost):
        """
        :title: sssd fails to release file descriptor on child
         logs after receiving hup
        :id: 3e28f453-fae8-4f52-82d0-757a5bdd0b06
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1544457
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        user = 'foo1@%s' % domain_name
        check_login_client(multihost, user, 'Secret123')
        time.sleep(2)
        ps_cmd = "mv /var/log/sssd/krb5_child.log  " \
                 "/var/log/sssd/krb5_child.log.old"
        multihost.client[0].run_command(ps_cmd)
        ps_cmd = "pgrep sssd"
        cmd = multihost.client[0].run_command(ps_cmd)
        sssd_pid = cmd.stdout_text.split('\n')[0]
        ps_cmd = f"/bin/kill -HUP {sssd_pid}"
        multihost.client[0].run_command(ps_cmd)
        check_login_client(multihost, user, 'Secret123')
        time.sleep(2)
        multihost.client[0].run_command(ps_cmd)
        for file in ['krb5_child.log', 'krb5_child.log.old']:
            ps_cmd = f"ls -l /var/log/sssd/{file}"
            cmd = multihost.client[0].run_command(ps_cmd)
            if f'/var/log/sssd/{file}' in cmd.stdout_text:
                status = 'PASS'
            else:
                status = 'FAIL'
        assert status == 'PASS'

    @staticmethod
    @pytest.mark.tier1
    def test_sssd_not_check_gss_spengo(multihost, backupsssdconf):
        """
        :title: krb5/fips: sssd does not properly check GSS-SPNEGO
        :id: 8ba5427e-8abe-44b9-adaa-878d2418b189
        :customerscenario: True
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1868054
        """
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        del_params = {'ldap_sasl_mech': 'GSSAPI'}
        client.sssd_conf('domain/%s' % domain_name,
                         del_params, action='delete')
        domain_params = {'ldap_sasl_mech': 'GSS-SPNEGO'}
        client.sssd_conf('domain/example1', domain_params)
        client.clear_sssd_cache()
        user = 'foo1@%s' % domain_name
        check_login_client(multihost, user, 'Secret123')
        ps_grep = "grep GSS /var/log/sssd/*.log"
        cmd = multihost.client[0].run_command(ps_grep)
        err_msg = "SPNEGO cannot find mechanisms to negotiate"
        assert err_msg not in cmd.stdout_text

    @staticmethod
    def test_fips_as_req(multihost):
        """
        :title: krb5/fips: verify sssd accepts only elisted fips approved types
        :id: c5ab16d5-8636-4f50-992b-aa0f05e1a9e5
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        user = 'foo2@%s' % domain_name
        ldap_host = multihost.master[0].sys_hostname
        pcapfile = '/tmp/krb1.pcap'
        tcpdump_cmd = 'tcpdump -s0 host %s -w %s' % (ldap_host, pcapfile)
        multihost.client[0].run_command(tcpdump_cmd, bg=True)
        pkill = 'pkill tcpdump'
        try:
            check_login_client(multihost, user, 'Secret123')
        except Exception:
            multihost.client[0].run_command(pkill)
            tshark_cmd = "tshark -r %s -V -2 -R 'kerberos.ENCTYPE'" % pcapfile
            multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
            pytest.fail("%s failed to login" % user)
        else:
            time.sleep(5)
            multihost.client[0].run_command(pkill)
            # check as_req
            tshark_cmd = "tshark -r %s -V -2 -R 'kerberos.ENCTYPE'" % pcapfile
            cmd = multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
            valid_etypes = ['AES128-CTS-HMAC-SHA256-128',
                            'AES256-CTS-HMAC-SHA1-96',
                            'AES128-CTS-HMAC-SHA1-96',
                            'AES256-CTS-HMAC-SHA384-192']
            for etype in valid_etypes:
                check = re.compile(r'%s' % etype)
                assert check.search(cmd.stdout_text)
        rm_pcap_file = 'rm -f %s' % pcapfile
        multihost.client[0].run_command(rm_pcap_file)

    @staticmethod
    def test_fips_as_rep(multihost):
        """
        :title: krb5/fips: verify sssd accepts only elisted fips approved types
        :id: f8452ecd-e13c-4485-83d3-83e25d7d544a
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        user = 'foo3@%s' % domain_name
        pcapfile = '/tmp/krb1.pcap'
        tcpdump_cmd = f'tcpdump -s0 host {multihost.master[0].sys_hostname}' \
                      f' -w {pcapfile}'
        multihost.client[0].run_command(tcpdump_cmd, bg=True)
        pkill = 'pkill tcpdump'
        try:
            check_login_client(multihost, user, 'Secret123')
        except Exception:
            multihost.client[0].run_command(pkill)
            print("SSH Login failed")
            tshark_cmd = "tshark -r %s -V -2 -R"\
                         " 'kerberos.msg_type == 11'" % pcapfile
            multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
            pytest.fail("%s failed to login" % user)
        else:
            time.sleep(5)
            multihost.client[0].run_command(pkill)
            # check as_rep
            tshark_cmd = "tshark -r %s -V -2 -R"\
                         " 'kerberos.msg_type == 11'" % pcapfile
            cmd = multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
            valid_etypes = ['AES256-CTS-HMAC-SHA1-96',
                            'AES128-CTS-HMAC-SHA1-96',
                            'AES256-CTS-HMAC-SHA384-192',
                            'AES128-CTS-HMAC-SHA256-128']
            count = 0
            for etype in valid_etypes:
                check = re.compile(r'%s' % etype)
                if check.search(cmd.stdout_text):
                    count += 1
        assert count == 1
        rm_pcap_file = 'rm -f %s' % pcapfile
        multihost.client[0].run_command(rm_pcap_file)

    @staticmethod
    def test_login_fips_weak_crypto(multihost):
        """
        :title: krb5/fips: verify login fails when weak crypto is presented
        :id: cdd2ef0d-4921-40b3-b61e-0b271b2d5e00
        """
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        tools.clear_sssd_cache()
        user = 'cracker@%s' % domain_name
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        krb = krb5srv(multihost.master[0], 'EXAMPLE.TEST')
        user_info = {'cn': 'cracker',
                     'uid': 'cracker',
                     'uidNumber': '19583100',
                     'gidNumber': '14564100'}
        if ldap_inst.posix_user("ou=People", "dc=example,dc=test", user_info):
            krb.add_principal('cracker', 'user', 'Secret123',
                              etype='arcfour-hmac')
        else:
            pytest.fail("Failed to add user cracker")
        user_dn = 'uid=cracker,ou=People,%s' % ds_suffix
        group_dn = 'cn=ldapusers,ou=Groups,%s' % ds_suffix
        add_member = [(ldap.MOD_ADD, 'uniqueMember', user_dn.encode('utf-8'))]
        (ret, _) = ldap_inst.modify_ldap(group_dn, add_member)
        assert ret == 'Success'
        tools.clear_sssd_cache()
        ldap_host = multihost.master[0].sys_hostname
        pcapfile = '/tmp/krb1.pcap'
        tcpdump_cmd = 'tcpdump -s0 host %s -w %s' % (ldap_host, pcapfile)
        multihost.client[0].run_command(tcpdump_cmd, bg=True)
        pkill = 'pkill tcpdump'
        try:
            check_login_client(multihost, user, 'Secret123')
        except Exception:
            multihost.client[0].run_command(pkill)
            tshark_cmd = "tshark -r %s -V -2 -R"\
                         " 'kerberos.msg_type == 30'" % pcapfile
            multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
            journalctl_cmd = 'journalctl --no-pager -n 150'
            cmd = multihost.client[0].run_command(journalctl_cmd)
            check = re.compile(r'KDC has no support for encryption type')
            assert check.search(cmd.stdout_text)
        else:
            pytest.fail("%s Login successfull")
        ldap_inst.del_dn(user_dn)
        krb.delete_principal('cracker')
        rm_pcap_file = 'rm -f %s' % pcapfile
        multihost.client[0].run_command(rm_pcap_file)

    @staticmethod
    def test_ldap_gssapi(multihost):
        """
        :title: krb5/fips: verify sssd is able to create gssapi connection
         with fips approved etype.
        :id: 8e80ddc7-fe6a-4729-91b6-f1fbae0dad73
        """
        cmd = 'cat /etc/sssd/sssd.conf'
        multihost.client[0].run_command(cmd)
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        user = 'foo1@%s' % domain_name
        ldap_host = multihost.master[0].sys_hostname
        pcapfile = '/tmp/ldapgssapi.pcap'
        tcpdump_cmd = 'tcpdump -s0 host %s -w %s' % (ldap_host, pcapfile)
        multihost.client[0].run_command(tcpdump_cmd, bg=True)
        pkill = 'pkill tcpdump'
        client = SSHClient(multihost.client[0].sys_hostname, user, 'Secret123')
        try:
            client.connect()
        except Exception:
            multihost.client[0].run_command(pkill)
            pytest.fail("%s failed to login" % user)
        else:
            ldapsearch = 'ldapsearch -Y GSSAPI -H ldap://%s' % ldap_host
            client.execute_command(ldapsearch)
            client.close()
            multihost.client[0].run_command(pkill, raiseonerr=False)
            tshark_cmd = "tshark -r %s -V -2 -R 'kerberos.msg_type == 13'" % pcapfile
            multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
        rm_pcap_file = 'rm -f %s' % pcapfile
        multihost.client[0].run_command(rm_pcap_file)

    @staticmethod
    def test_tgs_nonfips(multihost):
        """
        :title: krb5/fips: Verify sssd fails to create gssapi connection
         with weak etypes
        :id: f9623bc8-6305-45f1-ab6c-b9c2d18cdb8e
        """
        tools = sssdTools(multihost.client[0])
        host = multihost.client[0].sys_hostname
        domain_name = tools.get_domain_section_name()
        user = 'foo1@%s' % domain_name
        # stop the sssd service
        multihost.client[0].service_sssd('stop')
        add_principal = "add_principal -clearpolicy"\
                        " -e arcfour-hmac -randkey host/%s" % host
        cmd1 = "kadmin -p root/admin -w Secret123 -q '%s'" % add_principal
        multihost.client[0].run_command(cmd1, raiseonerr=False)
        ktadd = 'ktadd -k /etc/krb5.keytab -e arcfour-hmac host/%s' % host
        cmd2 = "kadmin -p root/admin -w Secret123 -q '%s'" % ktadd
        multihost.client[0].run_command(cmd2)
        klist = 'klist -kte /etc/krb5.keytab'
        multihost.client[0].run_command(klist)
        # start the sssd service
        tools.clear_sssd_cache()
        # verify ldap_child.log
        kdc_str = "KDC has no support for encryption type."\
                  " Unable to create GSSAPI-encrypted LDAP connection"
        check = re.compile(r'%s' % kdc_str)
        # check the ldap_child.login
        tail_cmd = 'tail -n 100 /var/log/sssd/ldap_child.log'
        cmd = multihost.client[0].run_command(tail_cmd, raiseonerr=False)
        if check.search(cmd.stdout_text):
            getent = 'getent passwd %s' % user
            cmd = multihost.client[0].run_command(getent, raiseonerr=False)
            assert cmd.returncode == 2
        else:
            print("'%s' message not seen on the logs" % kdc_str)



"""
SSSD Logging Tests.

client.sssd.start(debug_level=None), means no configuration. It is the same as if
the parameter is omitted from 'sssd.conf'.

:requirement: SSSD - Default debug level
"""

from __future__ import annotations

import time

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.topology import KnownTopology


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.Client)
def test_logging__default_settings_logs_debug_level(client: Client):
    """
    :title: Default settings writes the debug level to logs
    :setup:
        1. Configure SSSD for local system authentication
        2. Clear logs and start SSSD with default debug level
    :steps:
        1. Check log files
    :expectedresults:
        1. Logs messages contain default debug level 0x0070
    :customerscenario: False
    """
    client.sssd.common.local()
    client.sssd.default_domain = "local"

    client.sssd.clear(logs=True)
    client.sssd.start(debug_level=None)

    for file in [client.sssd.logs.monitor, client.sssd.logs.domain(), client.sssd.logs.nss, client.sssd.logs.pam]:
        log_str = client.fs.read(file)
        assert "level = 0x0070" in log_str, "Logs should contain debug_level = 0x0070!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.Client)
def test_logging__default_settings_does_not_log_user_logins(client: Client):
    """
    :title: Default debug level does not log user logins
    :setup:
        1. Create user
        2. Configure SSSD for local system authentication
        3. Clear cache and logs and start SSSD with default debug level
    :steps:
        1. Store current logs and authenticate as a local user.
        2. Compare stored logs with the current ones.
    :expectedresults:
        1. Login was successful
        2. Before event did not generate any new logs lines
    :customerscenario: False
    """

    client.local.user("user1").add()
    client.sssd.common.local()
    client.sssd.default_domain = "local"
    client.sssd.domain["fallback_homedir"] = "/home/%%u"

    client.sssd.clear(logs=True, config=False)
    client.sssd.start(debug_level=None)

    client.fs.copy("/var/log/sssd", "/tmp/copy")
    assert client.auth.ssh.password("user1", "Secret123"), "Login failed!"
    assert not client.host.conn.run("diff /var/log/sssd /tmp/copy").stdout, "Debug messages were generated!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.ticket(bz=1893159)
@pytest.mark.topology(KnownTopology.Client)
def test_logging__default_settings_logs_domain_configuration_errors(client: Client):
    """
    :title: Default debug_level logs domain configuration errors
    :setup:
        1. Configure SSSD with an invalid domain
    :steps:
        1. Start SSSD with default debug level
        2. Check logs
    :expectedresults:
        1. SSSD failed to start
        2. Logs contain error message
    :customerscenario: True
    """
    client.sssd.sssd["domains"] = "non_existing_domain"
    assert client.sssd.start(debug_level=None, raise_on_error=False).rc != 0, "SSSD erroneously started!"
    assert "No properly configured domains, fatal error!" in client.fs.read(
        client.sssd.logs.monitor
    ), "Domain is configured!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.ticket(bz=1893159)
@pytest.mark.topology(KnownTopology.LDAP)
def test_logging__default_settings_logs_offline_errors(client: Client):
    """
    :title: Default debug_level logs offline errors
    :setup:
        1. Configure SSSD with an invalid uri and enable ifp responder
        2. Start SSSD with default debug level
        3. Enable infopipe responder
    :steps:
        1. Check logs
        2. Check default domain status
    :expectedresults:
        1. Logs contain connection errors
        2. SSSD is not connected
    :customerscenario: True
    """
    client.sssd.domain["ldap_uri"] = "ldap://typo.invalid"
    client.sssd.enable_responder("ifp")
    client.sssd.start(debug_level=None, raise_on_error=False)

    logs = client.fs.read(client.sssd.logs.domain())
    assert "Failed to connect, going offline" in logs, "Offline error messages are not in logs!"

    assert client.sssd.default_domain is not None, "Failed to load default domain!"
    result = client.sssctl.domain_status(client.sssd.default_domain)
    assert result is not None
    assert "LDAP: not connected" in result.stdout, "LDAP is connected!"


@pytest.mark.integration
@pytest.mark.importance("low")
@pytest.mark.ticket(bz=1416150)
@pytest.mark.topology(KnownTopology.LDAP)
def test_logging__default_settings_logs_to_syslog_when_ldap_is_offline(client: Client):
    """
    :title: Log to syslog when sssd cannot contact ldap servers and the servers go offline
    :setup:
        1. Configure SSSD with an invalid uri and start SSSD
    :steps:
        1. Check domain status using sssctl
        2. Clear syslog and restart SSSD and check syslog
    :expectedresults:
        1. Domain is offline
        2. Logs contain SSSD errors
    :customerscenario: True
    """
    client.sssd.domain["ldap_uri"] = "ldaps://typo.invalid"
    client.sssd.start()

    assert client.sssd.default_domain is not None, "Failed to load default domain!"
    status = client.sssctl.domain_status(client.sssd.default_domain)
    assert status is not None
    assert "Offline" in status.stdout or "Unable to get online status" in status.stderr, "Domain is not offline!"

    client.journald.clear()
    client.sssd.restart()
    time.sleep(1)

    log = client.journald.journalctl(grep="Backend is offline", unit="sssd")
    assert log.rc == 0, "Offline error messages are not in logs!"



""" IPA AD Trust HBAC Cases

:requirement: HBAC (ipa_provider)
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import pytest
import re
from sssd.testlib.ipa.utils import ipaTools
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.exceptions import SSSDException


@pytest.mark.usefixtures('disable_allow_all_hbac',
                         'create_ad_users',
                         'create_ad_groups')
@pytest.mark.tier2
@pytest.mark.adhbac
class TestADTrustHbac(object):
    """ AD Trust HBAC Test cases """
    def test_allowed_ad_group(self, multihost):
        """
        :title: Verify Member of allowed AD Group
         through hbac is able to login
        :id: 401fb710-b876-4693-92d0-86a75b94973f
        """
        ipa_server_tools = ipaTools(multihost.master[0])
        ipa_client = sssdTools(multihost.client[0])
        ipa_server = sssdTools(multihost.master[0])
        client_host = multihost.client[0].sys_hostname
        ad_domain_name = multihost.ad[0].domainname.lower()
        aduser = 'idm_user1@%s' % ad_domain_name
        adgroup = 'idm_group1@%s' % ad_domain_name
        status = ''
        try:
            ipa_server_tools.create_group('idm_ext_group1', external=True)
        except SSSDException:
            status = 'FAIL'
        try:
            ipa_server_tools.group_add_member(adgroup, 'idm_ext_group1',
                                              external=True)
        except SSSDException:
            status = 'FAIL'

        try:
            ipa_server_tools.create_group('idm_posix_group1')
        except SSSDException:
            status = 'FAIL'
        try:
            ipa_server_tools.group_add_member('idm_ext_group1',
                                              'idm_posix_group1')
        except SSSDException:
            status = 'FAIL'
        ipa_server_tools.add_hbac_rule('ad_test1', 'idm_posix_group1',
                                       client_host, 'sshd', group=True)
        sssctl_cmd = 'sssctl user-checks -s sshd %s' % aduser
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
        cmd = multihost.client[0].run_command(sssctl_cmd, raiseonerr=False)
        result = test_pam.search(cmd.stderr_text)
        if not result:
            status = 'FAIL'
        else:
            status = 'PASS'
        for i in ['idm_ext_group1', 'idm_posix_group1']:
            cmd = 'ipa group-del %s' % i
            multihost.master[0].run_command(cmd, raiseonerr=False)
        ipa_server_tools.del_hbac_rule('ad_test1')
        ipa_client.clear_sssd_cache()
        ipa_server.clear_sssd_cache()
        assert status == 'PASS'

    def test_disallowed_ad_group(self, multihost, create_aduser_group):
        """
        :title: Verify Member of denied AD Group through
         hbac is not able to login
        :id: 7092f403-ca58-4683-89b8-400c64dd0a1d
        """
        (aduser, adgroup) = create_aduser_group
        ipa_server_tools = ipaTools(multihost.master[0])
        ipa_client = sssdTools(multihost.client[0])
        ipa_server = sssdTools(multihost.master[0])
        client_host = multihost.client[0].sys_hostname
        ad_domain_name = multihost.ad[0].domainname.lower()
        allow_adgroup = 'idm_group1@%s' % ad_domain_name
        status = ''
        try:
            ipa_server_tools.create_group('idm_ext_group2', external=True)
        except SSSDException:
            status = 'FAIL'
        try:
            ipa_server_tools.group_add_member(allow_adgroup, 'idm_ext_group2',
                                              external=True)
        except SSSDException:
            status = 'FAIL'

        try:
            ipa_server_tools.create_group('idm_posix_group2')
        except SSSDException:
            status = 'FAIL'
        try:
            ipa_server_tools.group_add_member('idm_ext_group2',
                                              'idm_posix_group2')
        except SSSDException:
            status = 'FAIL'
        ipa_server_tools.add_hbac_rule('ad_test2', 'idm_posix_group2',
                                       client_host, 'sshd', group=True)
        diallowed_user = '%s@%s' % (aduser, ad_domain_name)
        sssctl_cmd = 'sssctl user-checks -s sshd %s' % diallowed_user
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Permission denie')
        cmd = multihost.client[0].run_command(sssctl_cmd, raiseonerr=False)
        result = test_pam.search(cmd.stderr_text)
        if not result:
            status = 'FAIL'
        else:
            status = 'PASS'
        for i in ['idm_ext_group2', 'idm_posix_group2']:
            cmd = 'ipa group-del %s' % i
            multihost.master[0].run_command(cmd, raiseonerr=False)
        ipa_server_tools.del_hbac_rule('ad_test2')
        ipa_client.clear_sssd_cache()
        ipa_server.clear_sssd_cache()
        assert status == 'PASS'

    def test_multiple_ad_groups(self, multihost):
        """
        :title: Verify hbac evaluation when user is member
         of multiple AD Groups and with different hbac rules
        :id: eb78448d-8a4d-4800-9334-8d8cdb8b0af2
        """
        ipa_server_tools = ipaTools(multihost.master[0])
        ipa_client = sssdTools(multihost.client[0])
        ipa_server = sssdTools(multihost.master[0])
        client_host = multihost.client[0].sys_hostname
        ad_domain_name = multihost.ad[0].domainname.lower()
        aduser = 'idm_user3@%s' % ad_domain_name
        adgroup = 'idm_group3@%s' % ad_domain_name
        status = ''
        for i in range(3, 5, 1):
            ext_group = 'idm_ext_group%d' % i
            adgroup = 'idm_group%d@%s' % (i, ad_domain_name)
            posix_group = 'idm_posix_group%d' % i
            try:
                ipa_server_tools.create_group(ext_group, external=True)
            except SSSDException:
                status = 'FAIL'
            try:
                ipa_server_tools.group_add_member(adgroup, ext_group,
                                                  external=True)
            except SSSDException:
                status = 'FAIL'
            try:
                ipa_server_tools.create_group(posix_group)
            except SSSDException:
                status = 'FAIL'
            try:
                ipa_server_tools.group_add_member(ext_group, posix_group)
            except SSSDException:
                status = 'FAIL'

        ipa_server_tools.add_hbac_rule('ad_test3', 'idm_posix_group3',
                                       client_host, 'sshd', group=True)
        ipa_server_tools.add_hbac_rule('ad_test4', 'idm_posix_group4',
                                       client_host, 'sudo', group=True)
        sssctl_cmd = 'sssctl user-checks -s sshd %s' % aduser
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
        cmd = multihost.client[0].run_command(sssctl_cmd, raiseonerr=False)
        result = test_pam.search(cmd.stderr_text)
        if not result:
            status = 'FAIL'
        else:
            status = 'PASS'
        for i in ['idm_ext_group3', 'idm_ext_group4',
                  'idm_posix_group3', 'idm_posix_group4']:
            cmd = 'ipa group-del %s' % i
            multihost.master[0].run_command(cmd, raiseonerr=False)
        ipa_server_tools.del_hbac_rule('ad_test3')
        ipa_server_tools.del_hbac_rule('ad_test4')
        ipa_client.clear_sssd_cache()
        ipa_server.clear_sssd_cache()
        assert status == 'PASS'

    def test_hbac_nested_group(self, multihost):
        """
        :title: Verify hbac evaluation of AD Nested Groups
        :id: f7fc6349-daba-43c2-be4e-e13923e201f9
        """
        ipa_server_tools = ipaTools(multihost.master[0])
        ipa_client = sssdTools(multihost.client[0])
        ipa_server = sssdTools(multihost.master[0])
        client_host = multihost.client[0].sys_hostname
        ad_domain_name = multihost.ad[0].domainname.lower()
        aduser = 'idm_user1@%s' % ad_domain_name
        adgroup = 'nested_group1@%s' % ad_domain_name
        status = ''
        try:
            ipa_server_tools.create_group('idm_ext_group5', external=True)
        except SSSDException:
            status = 'FAIL'
        try:
            ipa_server_tools.group_add_member(adgroup, 'idm_ext_group5',
                                              external=True)
        except SSSDException:
            status = 'FAIL'

        try:
            ipa_server_tools.create_group('idm_posix_group5')
        except SSSDException:
            status = 'FAIL'
        try:
            ipa_server_tools.group_add_member('idm_ext_group5',
                                              'idm_posix_group5')
        except SSSDException:
            status = 'FAIL'
        ipa_server_tools.add_hbac_rule('ad_test5', 'idm_posix_group5',
                                       client_host, 'sshd', group=True)
        sssctl_cmd = 'sssctl user-checks -s sshd %s' % aduser
        test_pam = re.compile(r'%s' % 'pam_acct_mgmt: Success')
        cmd = multihost.client[0].run_command(sssctl_cmd, raiseonerr=False)
        result = test_pam.search(cmd.stderr_text)
        if not result:
            status = 'FAIL'
        else:
            status = 'PASS'
        for i in ['idm_ext_group5', 'idm_posix_group5']:
            cmd = 'ipa group-del %s' % i
            multihost.master[0].run_command(cmd, raiseonerr=False)
        ipa_server_tools.del_hbac_rule('ad_test5')
        ipa_client.clear_sssd_cache()
        ipa_server.clear_sssd_cache()
        assert status == 'PASS'



""" test cases for sssd proxy

:requirement: IDM-SSSD-REQ : Proxy Provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""


import textwrap
import time
from string import Template
import pytest
from constants import ds_suffix
from sssd.testlib.common.ssh2_python import SSHClient
from sssd.testlib.common.utils import sssdTools


@pytest.mark.usefixtures('setup_sssd_krb', 'create_host_keytab',
                         'create_posix_usersgroups')
@pytest.mark.proxy
class TestsssdProxy(object):
    """ Testing sssd-proxy  """
    @pytest.mark.tier1
    def test_0001_1724717(self, multihost):
        """
        :title: proxy: sssd-proxy crashes resolving groups with no members
        :id: 28b64673-8f1b-46c1-b0dd-6eaba9f80b2c
        """
        # backup sssd.conf
        backup = 'cp -f /etc/sssd/sssd.conf /etc/sssd/sssd.conf.backup'
        restore = 'cp -f /etc/sssd/sssd.conf.backup /etc/sssd/sssd.conf'
        multihost.client[0].run_command(backup)
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        user = 'foo1@%s' % domain_name
        # user add
        add_user = 'useradd foo1'
        # delete user
        del_user = 'userdel -rf foo1'
        multihost.client[0].run_command(add_user)
        domain_params = {'id_provider': 'proxy',
                         'proxy_lib_name': 'files',
                         'ignore_group_members': 'False',
                         'cache_credentials': 'True',
                         'krb5_validate': 'True'}
        tools.sssd_conf('domain/%s' % domain_name, domain_params)
        del_domain_params = {'ldap_uri': 'ldaps:%s' %
                             (multihost.master[0].run_command),
                             'ldap_tls_cacert':
                             '/etc/openldap/cacerts/cacert.pem',
                             'ldap_search_base': ds_suffix,
                             'use_fully_qualified_names': 'True'}
        tools.sssd_conf('domain/%s' % domain_name,
                        del_domain_params, action='delete')
        cat = 'cat /etc/sssd/sssd.conf'
        multihost.client[0].run_command(cat)
        multihost.client[0].service_sssd('start')
        client = SSHClient(multihost.client[0].sys_hostname, user, 'Secret123')
        try:
            client.connect()
        except Exception:
            multihost.client[0].run_command(del_user)
            multihost.client[0].run_command(restore)
            pytest.fail("%s failed to login" % user)
        else:
            id_cmd = 'id %s' % user
            ret1 = client.execute_command(id_cmd)
            assert "no such user" not in ret1
            client.close()
        # On fedora after user logs out it takes time
        # for systemd process running as user to get stopped, hence
        # adding sleep
        time.sleep(20)
        multihost.client[0].run_command(del_user)
        multihost.client[0].run_command(restore)

    def test_0003_update_removed_grp_membership(self, multihost,
                                                backupsssdconf):
        """
        :title: proxy: secondary group is shown in sssd cache after
         group is removed
        :id: 7cfb9aa9-6e68-4914-afb8-ecfae132aa84
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1917970
        :customerscenario: true
        :steps:
          1. Edit sssd.conf and configure proxy provider with
             entry_cache_timeout = 1
          2. Restart SSSD with cleared cache
          3. Create a localuser and localgroup
          4. Add that localuser to the localgroup
          5. Assert localgroup is shown in localuser's group list
          6. Remove localuser from localgroup
          7. Assert that after entry_cache_timeout, localuser's groups
             are not listing localgroup
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          7. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        l_usr, l_grp = 'testuser', 'testgroup'
        multihost.client[0].run_command(f'useradd {l_usr}')
        multihost.client[0].run_command(f'groupadd {l_grp}')
        multihost.client[0].run_command(f'usermod -aG {l_grp} {l_usr}')
        domain_params = {'id_provider': 'proxy',
                         'proxy_lib_name': 'files',
                         'auth_provider': 'krb5',
                         'ignore_group_members': 'False',
                         'cache_credentials': 'True',
                         'entry_cache_timeout': '1',
                         'krb5_validate': 'True'}
        tools.sssd_conf('domain/%s' % domain_name, domain_params)
        del_domain_params = {'ldap_uri': 'ldaps:%s' %
                             (multihost.master[0].run_command),
                             'ldap_tls_cacert':
                             '/etc/openldap/cacerts/cacert.pem',
                             'ldap_search_base': ds_suffix,
                             'use_fully_qualified_names': 'True'}
        tools.sssd_conf('domain/%s' % domain_name,
                        del_domain_params, action='delete')
        tools.clear_sssd_cache()
        cmd = multihost.client[0].run_command(f'groups {l_usr}')
        assert 'testgroup' in cmd.stdout_text
        multihost.client[0].run_command(f'gpasswd -d {l_usr} {l_grp}')
        time.sleep(1)
        cmd = multihost.client[0].run_command(f'groups {l_usr}')
        multihost.client[0].run_command(f'userdel -rf {l_usr}')
        multihost.client[0].run_command(f'groupdel -f {l_grp}')
        assert 'testgroup' not in cmd.stdout_text

    def test_innetgr_threads(self, multihost, backupsssdconf):
        """
        :title: Verify sssd is thread-safe in innetgr
        :id: d38a8279-312d-4f52-808c-17226e7168d3
        :customerscenario: True
        :description:
         SSSD was not thread safe in innetgr call when using nested netgroups
         resulting in nfs-ganesha not working correctly
        :setup:
          1. Configure client to use sssd proxy files
          2. Create /etc/netgroup file with two groups containing 1000+
             members.
          3. Add the group/host info to the c sources
          4. Compile the sources using gcc
        :steps:
          1. Restart sssd on client and clear caches
          2. Run first binary to verify that the setup is correct.
          3. Run second binary to verify that the bug is fixed.
        :expectedresults:
          1. SSSD restarted successfully
          2. Test binary returns 0 return code.
          3. Test binary returns 0 return code.
        :teardown:
          1. Remove the net groups
          2. Remove the sources and binaries.
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1703436
        """
        # SETUP
        # Configure proxy provider
        sssd_client = sssdTools(multihost.client[0])
        domain_params = {'id_provider': 'proxy',
                         'proxy_lib_name': 'files',
                         'auth_provider': 'none'
                         }
        sssd_client.sssd_conf(
            'domain/%s' % sssd_client.get_domain_section_name(),
            domain_params
        )

        # Create the net groups in file
        client_shortname = multihost.client[0].shortname
        net_group1_name = "ngr"
        net_group1 = f"{net_group1_name} "
        net_group2 = f"{net_group1_name}2 "

        # We need long enough member searching so the issue appears.
        # This number of members reproduced failure in 2+ threads reliably.
        for i in range(1, 4096):
            net_group1 += "(host1%04d, user1%04d, domain1) " % (i, i)
            net_group2 += "(host2%04d, user2%04d, domain2) " % (i, i)
        net_group1 += f"({client_shortname}, myuser, domain6)"

        multihost.client[0].transport.put_file_contents(
            '/etc/netgroup',
            net_group1 + "\n" + net_group2 + "\n"
        )

        # Prepare c code
        pre = "pthread_mutex_lock(&netg_lock);"
        post = "pthread_mutex_unlock(&netg_lock);"
        code_template_str = textwrap.dedent("""\
            #include <stdio.h>
            #include <stdlib.h>
            #include <netdb.h>
            #include <pthread.h>
            #define NTHREADS 256
            #define NUM_CALLS 1000
            static char *groups[] = {"$group"};
            static char *hosts[] = {"$host"};
            #define NHOSTS (sizeof(hosts)/sizeof(hosts[0]))
            #define NGROUPS (sizeof(groups)/sizeof(groups[0]))
            static int pass_count;
            static int fail_count;
            static int bogus_count;
            static pthread_mutex_t netg_lock = PTHREAD_MUTEX_INITIALIZER;
            static void *thread_main(void *arg)
            {
                unsigned long i;
                char *host;
                char *group;
                int rc;
                for (i = 0; i < NUM_CALLS; i++) {
                    host = hosts[rand() % NHOSTS];
                    group = groups[rand() % NGROUPS];
                    $pre
                    rc = innetgr(group, host, NULL, NULL);
                    $post
                    /* Ideally, atomic increments should be used, but
                     * rough numbers are OK for now
                     */
                    if (rc == 0)
                        fail_count++;
                    else if (rc == 1)
                        pass_count++;
                    else
                        bogus_count++;
                }
            }
            int main()
            {
                pthread_t threads[NTHREADS];
                int i;
                for (i = 0; i < NTHREADS; i++)
                    pthread_create(&threads[i], NULL, thread_main, (void *)0);
                for (i = 0; i < NTHREADS; i++)
                    pthread_join(threads[i], NULL);
                printf("pass:%d, fail:%d, bogus:%d\\n",
                        pass_count, fail_count, bogus_count);
                if ( fail_count > 0 )
                    exit(2);
            }""")
        code_template = Template(code_template_str)

        # Substitute c code and upload
        code = code_template.substitute(
            group=net_group1_name,
            host=client_shortname,
            pre=pre,
            post=post
        )
        multihost.client[0].transport.put_file_contents('/root/netg-lock.c',
                                                        code)
        code = code_template.substitute(
            group=net_group1_name,
            host=client_shortname,
            pre='',
            post=''
        )
        multihost.client[0].transport.put_file_contents('/root/netg-lock2.c',
                                                        code)

        # Install packages including gcc
        sssdTools(multihost.client[0]).client_install_pkgs()

        # Compile c code
        compile_cmd = 'gcc -lpthread -o /root/netg-lock /root/netg-lock.c ' \
                      '&& gcc -lpthread -o /root/netg-lock2 /root/netg-lock2.c'
        gcc = multihost.client[0].run_command(compile_cmd, raiseonerr=False)
        chmod_cmd = 'chmod +x /root/netg-lock*'
        multihost.client[0].run_command(chmod_cmd, raiseonerr=False)

        # TEST EXECUTION
        sssd_client.clear_sssd_cache()
        cmd1 = multihost.client[0].run_command('/root/netg-lock',
                                               raiseonerr=False)
        cmd2 = multihost.client[0].run_command('/root/netg-lock2',
                                               raiseonerr=False)

        # TEARDOWN
        multihost.client[0].run_command(
            'rm -f /root/netg-lock*; rm -f /etc/netgroup',
            raiseonerr=False
        )

        # TEST EVALUATION
        assert gcc.returncode == 0, 'Compiling of binaries failed!'
        assert cmd1.returncode == 0, 'First binary failed, incorrect setup!'
        assert cmd2.returncode == 0, 'Second binary failed, test failed!'



#
# Resolver integration test
#
#   Authors:
#       Samuel Cabrero <scabrero@suse.com>
#
#   Copyright (C) 2019 SUSE LINUX GmbH, Nuernberg, Germany.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import stat
import signal
import subprocess
import time
import ldap
import pytest
import socket
import config
import ds_openldap
import ldap_ent
from util import unindent
from sssd_nss import NssReturnCode, HostError
from sssd_hosts import call_sssd_gethostbyname
from sssd_nets import call_sssd_getnetbyname, call_sssd_getnetbyaddr

LDAP_BASE_DN = "dc=example,dc=com"


@pytest.fixture(scope="module")
def ds_inst(request):
    """LDAP server instance fixture"""
    ds_inst = ds_openldap.DSOpenLDAP(
        config.PREFIX, 10389, LDAP_BASE_DN,
        "cn=admin", "Secret123"
    )

    try:
        ds_inst.setup()
    except Exception:
        ds_inst.teardown()
        raise
    request.addfinalizer(ds_inst.teardown)
    return ds_inst


@pytest.fixture(scope="module")
def ldap_conn(request, ds_inst):
    """LDAP server connection fixture"""
    ldap_conn = ds_inst.bind()
    ldap_conn.ds_inst = ds_inst
    request.addfinalizer(ldap_conn.unbind_s)
    return ldap_conn


def create_ldap_entries(ldap_conn, ent_list=None):
    """Add LDAP entries from ent_list"""
    if ent_list is not None:
        for entry in ent_list:
            ldap_conn.add_s(entry[0], entry[1])


def cleanup_ldap_entries(ldap_conn, ent_list=None):
    """Remove LDAP entries added by create_ldap_entries"""
    if ent_list is None:
        for ou in ("Users", "Groups", "Netgroups", "Services", "Policies",
                   "Hosts", "Networks"):
            for entry in ldap_conn.search_s(f"ou={ou},"
                                            f"{ldap_conn.ds_inst.base_dn}",
                                            ldap.SCOPE_ONELEVEL,
                                            attrlist=[]):
                ldap_conn.delete_s(entry[0])
    else:
        for entry in ent_list:
            ldap_conn.delete_s(entry[0])


def create_ldap_cleanup(request, ldap_conn, ent_list=None):
    """Add teardown for removing all user/group LDAP entries"""
    request.addfinalizer(lambda: cleanup_ldap_entries(ldap_conn, ent_list))


def create_ldap_fixture(request, ldap_conn, ent_list=None):
    """Add LDAP entries and add teardown for removing them"""
    create_ldap_entries(ldap_conn, ent_list)
    create_ldap_cleanup(request, ldap_conn, ent_list)


SCHEMA_RFC2307 = "rfc2307"
SCHEMA_RFC2307_BIS = "rfc2307bis"


def format_basic_conf(ldap_conn, schema):
    """Format a basic SSSD configuration"""
    schema_conf = "ldap_schema         = " + schema + "\n"
    if schema == SCHEMA_RFC2307_BIS:
        schema_conf += "ldap_group_object_class = groupOfNames\n"
    iphost_search_base = "ou=Hosts," + ldap_conn.ds_inst.base_dn
    ipnetwork_search_base = "ou=Networks," + ldap_conn.ds_inst.base_dn
    return unindent("""\
        [sssd]
        debug_level         = 0xffff
        domains             = LDAP
        services            = nss

        [nss]
        debug_level         = 0xffff
        memcache_timeout    = 0
        entry_negative_timeout = 1

        [domain/LDAP]
        ldap_auth_disable_tls_never_use_in_production = true
        ldap_id_use_start_tls = false
        debug_level         = 0xffff
        {schema_conf}
        id_provider         = ldap
        auth_provider       = ldap
        resolver_provider   = ldap
        ldap_uri            = {ldap_conn.ds_inst.ldap_url}
        ldap_search_base    = {ldap_conn.ds_inst.base_dn}
        ldap_iphost_search_base = {iphost_search_base}
        ldap_ipnetwork_search_base = {ipnetwork_search_base}
    """).format(**locals())


def create_conf_file(contents):
    """Create sssd.conf with specified contents"""
    conf = open(config.CONF_PATH, "w")
    conf.write(contents)
    conf.close()
    os.chmod(config.CONF_PATH, stat.S_IRUSR | stat.S_IWUSR)


def cleanup_conf_file():
    """Remove sssd.conf, if it exists"""
    if os.path.lexists(config.CONF_PATH):
        os.unlink(config.CONF_PATH)


def create_conf_cleanup(request):
    """Add teardown for removing sssd.conf"""
    request.addfinalizer(cleanup_conf_file)


def create_conf_fixture(request, contents):
    """
    Create sssd.conf with specified contents and add teardown for removing it
    """
    create_conf_file(contents)
    create_conf_cleanup(request)


def create_sssd_process():
    """Start the SSSD process"""
    if subprocess.call(["sssd", "-D", "--logger=files"]) != 0:
        raise Exception("sssd start failed")


def get_sssd_pid():
    pid_file = open(config.PIDFILE_PATH, "r")
    pid = int(pid_file.read())
    return pid


def cleanup_sssd_process():
    """Stop the SSSD process and remove its state"""
    try:
        pid = get_sssd_pid()
        os.kill(pid, signal.SIGTERM)
        while True:
            try:
                os.kill(pid, signal.SIGCONT)
            except OSError:
                break
            time.sleep(1)
    except OSError:
        pass
    for path in os.listdir(config.DB_PATH):
        os.unlink(config.DB_PATH + "/" + path)
    for path in os.listdir(config.MCACHE_PATH):
        os.unlink(config.MCACHE_PATH + "/" + path)


def create_sssd_fixture(request):
    """Start SSSD and add teardown for stopping it and removing its state"""
    create_sssd_process()
    create_sssd_cleanup(request)


def create_sssd_cleanup(request):
    """Add teardown for stopping SSSD and removing its state"""
    request.addfinalizer(cleanup_sssd_process)


@pytest.fixture
def add_hosts(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_host("host1",
                      aliases=["host1_alias1", "host1_alias2"],
                      addresses=["192.168.1.1", "192.168.1.2",
                                 "2001:db8:1::1", "2001:db8:1::2"])
    ent_list.add_host("host2.example.com",
                      aliases=["host2_alias1.example.com",
                               "host2_alias2.example.com"],
                      addresses=["192.168.2.1", "192.168.2.2",
                                 "2001:db8:2::1", "2001:db8:2::2"])

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


@pytest.fixture
def add_nets(request, ldap_conn):
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)

    ent_list.add_ipnet("net1", "192.168.1.1",
                       aliases=["net1_alias1", "net1_alias2"])
    ent_list.add_ipnet("net2", "10.2.2.2",
                       aliases=["net2_alias1", "net2_alias2"])

    create_ldap_fixture(request, ldap_conn, ent_list)
    conf = format_basic_conf(ldap_conn, SCHEMA_RFC2307)
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)
    return None


def test_hostbyname(add_hosts):
    (res, hres, _) = call_sssd_gethostbyname("invalid")

    assert res == NssReturnCode.NOTFOUND
    assert hres == HostError.HOST_NOT_FOUND

    (res, hres, _) = call_sssd_gethostbyname("example.com")
    assert res == NssReturnCode.NOTFOUND
    assert hres == HostError.HOST_NOT_FOUND

    (res, hres, _) = call_sssd_gethostbyname("invalid.example.com")
    assert res == NssReturnCode.NOTFOUND
    assert hres == HostError.HOST_NOT_FOUND

    (res, hres, _) = call_sssd_gethostbyname("host1")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_gethostbyname("host1_alias1")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_gethostbyname("host1_alias2")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_gethostbyname("host2.example.com")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_gethostbyname("host2_alias1.example.com")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_gethostbyname("host2_alias2.example.com")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0


def test_netbyname(add_nets):
    (res, hres, _) = call_sssd_getnetbyname("invalid")
    assert res == NssReturnCode.NOTFOUND
    assert hres == HostError.HOST_NOT_FOUND

    (res, hres, _) = call_sssd_getnetbyname("net1")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_getnetbyname("net1_alias1")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_getnetbyname("net1_alias2")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_getnetbyname("net2")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_getnetbyname("net2_alias1")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_getnetbyname("net2_alias2")
    assert res == NssReturnCode.SUCCESS
    assert hres == 0


def test_netbyaddr(add_nets):
    (res, hres, _) = call_sssd_getnetbyaddr("10.2.2.1", socket.AF_INET)
    assert res == NssReturnCode.NOTFOUND
    assert hres == HostError.HOST_NOT_FOUND

    (res, hres, _) = call_sssd_getnetbyaddr("10.2.2.1", socket.AF_UNSPEC)
    assert res == NssReturnCode.NOTFOUND
    assert hres == HostError.HOST_NOT_FOUND

    (res, hres, _) = call_sssd_getnetbyaddr("10.2.2.2", socket.AF_INET)
    assert res == NssReturnCode.SUCCESS
    assert hres == 0

    (res, hres, _) = call_sssd_getnetbyaddr("10.2.2.2", socket.AF_UNSPEC)
    assert res == NssReturnCode.SUCCESS
    assert hres == 0



"""
Tools Tests.

Tests pertaining to command line tools, some tools will have their own file.

* sssctl: test_sssctl.py
* sss_cache
* sss_obfuscate
* sss_seed
* sss_debuglevel
* sss_override: sss_override.py
* sss_ssh_authorizedkeys
* sss_ssh_knownhostsproxy

:requirement: Tools
"""

from __future__ import annotations

import pytest
from pytest_mh.conn import ProcessError
from sssd_test_framework.roles.client import Client
from sssd_test_framework.topology import KnownTopology


@pytest.mark.importance("medium")
@pytest.mark.ticket(bz=1661182)
@pytest.mark.topology(KnownTopology.Client)
def test_tools__sss_cache_expired_does_not_print_unrelated_message(client: Client):
    """
    :title: Usermod command does not print unrelated sss_cache messages
    :setup:
        1. Configure SSSD without any domain
        2. Set to sssd section "enable_files_domain" to "false"
        3. Create local user
    :steps:
        1. Restart SSSD
        2. Modify existing local user
        3. Expire cache with specific options
    :expectedresults:
        1. Error is raised, SSSD is not running
        2. Modified successfully
        3. Output did not contain wrong messages
    :customerscenario: True
    """
    client.sssd.sssd["enable_files_domain"] = "false"
    client.local.user("user1").add()

    with pytest.raises(ProcessError):
        client.sssd.restart()

    res = client.host.conn.run("usermod -a -G wheel user1")
    assert (
        "No domains configured, fatal error!" not in res.stdout
    ), "'No domains configured, fatal error!' printed to stdout!"

    for cmd in ("sss_cache -U", "sss_cache -G", "sss_cache -E", "sss_cache --user=nonexisting"):
        res = client.host.conn.run(cmd)
        assert (
            "No domains configured, fatal error!" not in res.stdout
        ), "'No domains configured, fatal error!' printed to stdout!"



""" This module defines classes regarding
functions/methods related to samba/winbind """

import os
import time
import subprocess
import pytest
import tempfile
from .utils import ADOperations
from .utils import sssdTools
from .paths import SMB_DEFAULT_CONF
from .exceptions import SSSDException


class sambaTools(object):
    """ Functions related to samba/winbind setup """
    def __init__(self, host, adhost):
        """ Initialize host """
        self.host = host
        self.adhost = adhost
        self.adhost_ip = self.adhost.ip
        self.adhost_ops = ADOperations(self.adhost)
        self._ad_conn = None
        self.adhost_realm = self.adhost.realm
        self.adhost_password = self.adhost.ssh_password
        self.adhost_hostname = self.adhost.external_hostname
        self.adhost_basedn = self.adhost.domain_basedn_entry
        self.adhost_adminuser = 'Administrator'
        self.host_tools = sssdTools(self.host, self.adhost)

    @property
    def adhost_conn(self):
        """AD LDAP connection in lazy initialized property"""
        if self._ad_conn:
            return self._ad_conn
        self._ad_conn = self.adhost_ops.ad_conn()
        return self._ad_conn

    def smbadsconf(self):
        """ Setup smbconf with security = ads """
        bkup = 'cp -a /etc/samba/smb.conf /etc/samba/smb.conf.orig'
        self.host.run_command(bkup, raiseonerr=False)
        client_hostname = self.host.sys_hostname
        client_short_name = client_hostname.strip().split('.')[0]
        tmpconf = tempfile.NamedTemporaryFile(suffix='smb.conf', delete=False)
        workgroup = self.adhost_realm.split('.')[0]
        global_parameters = {'workgroup': workgroup,
                             'security': 'ads',
                             'realm': self.adhost_realm}
        if len(client_short_name) > 15:
            client_short_name = client_short_name[:15]
            global_parameters['netbios name'] = client_short_name
        global_parameters['kerberos method'] = "secrets and keytab"
        global_parameters['client signing'] = "yes"
        global_parameters['client use spnego'] = "yes"
        global_parameters['log file'] = "/var/log/samba/log.%m"
        global_parameters['max log size'] = "50"
        global_parameters['log level'] = "9"
        try:
            self.host_tools.update_conf(tmpconf.name,
                                        'global',
                                        global_parameters)
        except SSSDException:
            raise
        else:
            self.host.transport.put_file(tmpconf.name, '/etc/samba/smb.conf')
            os.unlink(tmpconf.name)

    def enable_idmapsss(self, idmap_range=None, tdb_range=None):
        """ Enable sssd backend for idmap """
        tmpconf = tempfile.NamedTemporaryFile(suffix='smb.conf', delete=False)
        self.host.transport.get_file(SMB_DEFAULT_CONF, tmpconf.name)
        netbiosname = self.adhost.netbiosname.strip()
        idmap_backend = "idmap config %s : backend" % netbiosname
        idmap_sss_range = "idmap config %s : range" % netbiosname
        idmap_tdb = "idmap config * : backend"
        idmap_tdb_range = "idmap config * : range"
        if not idmap_range:
            idmap_range = '200000-2147483647'
        if not tdb_range:
            tdb_range = '50000-100000'

        idmap_params = {idmap_tdb: 'tdb',
                        idmap_tdb_range: tdb_range,
                        idmap_backend: 'sss',
                        idmap_sss_range: idmap_range}
        try:
            self.host_tools.update_conf(tmpconf.name, 'global',
                                        idmap_params, action='update')
        except SSSDException:
            raise
        else:
            self.host.transport.put_file(tmpconf.name, SMB_DEFAULT_CONF)
            os.unlink(tmpconf.name)

    def add_share_definition(self, share_name, share_path):
        """ Add samba share in smb.conf """
        tmpconf = tempfile.NamedTemporaryFile(suffix='smb.conf', delete=False)
        self.host.transport.get_file(SMB_DEFAULT_CONF, tmpconf.name)
        share_params = {'path': share_path,
                        'comment': 'test share %s' % share_name,
                        'writable': 'yes',
                        'printable': 'no'}
        try:
            self.host_tools.update_conf(tmpconf.name, 'share1',
                                        share_params, action='update')
        except SSSDException:
            raise
        else:
            self.host.transport.put_file(tmpconf.name, SMB_DEFAULT_CONF)
            os.unlink(tmpconf.name)

    def service_smb(self, action='start'):
        """ Start smb servicer """
        cmd = 'systemctl %s smb' % action
        self.host.run_command(cmd)

    def create_samba_share(self, share_path):
        """ Create samba share directory """
        create_share_dir = 'mkdir -p %s' % share_path
        self.host.run_command(create_share_dir)
        selinux_context = 'chcon -t samba_share_t -R %s' % share_path
        self.host.run_command(selinux_context)
        chgrp = "chgrp 'Domain Users@%s' -R %s" % (self.adhost_realm,
                                                   share_path)
        self.host.run_command(chgrp)
        chmod = 'chmod 770 %s' % share_path
        self.host.run_command(chmod)

    def delete_samba_share(self, share_path):
        """ Delete samba share directory """
        delete_share = 'rm -rf %s' % share_path
        self.host.run_command(delete_share)

    def enable_winbind(self):
        """ Enable winbind with idmap sss backend """
        # disjoin system first
        self.host_tools.update_resolv_conf(self.adhost_ip)
        self.host_tools.create_kdcinfo(self.adhost_realm, self.adhost_ip)
        self.host_tools.systemsssdauth(self.adhost_realm, self.adhost_hostname)
        self.host_tools.disjoin_ad(raiseonerr=False)
        kinit = "kinit Administrator"
        try:
            self.host.run_command(kinit, stdin_text=self.adhost_password)
        except subprocess.CalledProcessError:
            pytest.fail("kinit failed")
        self.host_tools.join_ad(self.adhost_realm, self.adhost_password,
                                mem_sw='samba')
        self.host_tools.service_ctrl("restart", "sssd")
        self.smbadsconf()
        self.enable_idmapsss()
        restart_winbind = 'systemctl restart winbind'
        cmd = self.host.run_command(restart_winbind, raiseonerr=False)
        assert cmd.returncode == 0
        time.sleep(20)

    def disable_winbind(self):
        """ Disable winbind """
        # disjoin system
        self.host_tools.disjoin_ad(raiseonerr=False)
        # stop winbind
        stop_winbind = 'systemctl stop winbind'
        self.host.run_command(stop_winbind)

    def remove_smbconf(self):
        """ Remove smb.conf """
        # remove smb.conf
        cmd = 'rm -f /etc/samba/smb.conf'
        self.host.run_command(cmd)

    def clear_samba_cache(self):
        """ Clear samba cache """
        cache_files = ['group_mapping.tdb',
                       'account_policy.tdb',
                       'registry.tdb',
                       'share_info.tdb',
                       'smbprofile.tdb',
                       'winbindd_idmap.tdb',
                       'netsamlogon_cache.tdb',
                       'winbindd_cache.tdb',
                       'private/netlogon_creds_cli.tdb',
                       'private/secrets.ldb',
                       'private/secrets.tdb',
                       'private/passdb.tdb']
        cache_path = '/var/lib/samba'
        for cache in cache_files:
            cmd = 'rm -f %s/%s' % (cache_path, cache)
            self.host.run_command(cmd)



"""Automation for Krb Fast Principal tests ported from bash

:requirement: krb_fast_principal
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

import pytest
import time
from constants import ds_instance_name
from sssd.testlib.common.ssh2_python import check_login_client_bool, check_login_client
from sssd.testlib.common.utils import sssdTools


@pytest.fixture(scope='class')
def create_keytab(multihost, request):
    """
    Configure keytab for this test
    """
    client_hostname = multihost.client[0].sys_hostname
    master_hostname = multihost.master[0].sys_hostname
    multihost.master[0].run_command("cp -vfr /opt /tmp/")
    multihost.client[0].run_command("cp -vfr /opt /tmp/")
    multihost.client[0].run_command("cp -vf /etc/krb5.keytab /etc/krb5.keytab_anuj")
    multihost.master[0].run_command("rm -vf /opt/*")
    multihost.client[0].run_command("rm -vf /opt/*")
    for command in [f'kadmin.local -q "addprinc -randkey ldap/{master_hostname}@EXAMPLE.TEST"',
                    f'kadmin.local -q "ktadd -k /etc/krb5.keytab ldap/{master_hostname}@EXAMPLE.TEST"',
                    f'kadmin.local -q "addprinc -randkey host/{client_hostname}@EXAMPLE.TEST"',
                    f'kadmin.local -q "ktadd -k /opt/sssd_client_invalid.keytab host/{client_hostname}@EXAMPLE.TEST"',
                    f'kadmin.local -q "ktadd -k /opt/sssd_client_valid.keytab host/{client_hostname}@EXAMPLE.TEST"',
                    'kadmin.local -q "addprinc -randkey host/invalidhost.redhat.com@EXAMPLE.TEST"',
                    'kadmin.local -q "ktadd -k /opt/invalid.keytab host/invalidhost.redhat.com@EXAMPLE.TEST"',
                    'kadmin.local -q "delprinc -force host/invalidhost.redhat.com@EXAMPLE.TEST"',
                    'chmod 777 /etc/krb5.keytab',
                    'restorecon -v /etc/krb5.keytab']:
        multihost.master[0].run_command(command, raiseonerr=False)
    file_content = multihost.master[0].get_file_contents("/opt/sssd_client_valid.keytab")
    multihost.client[0].put_file_contents("/etc/krb5.keytab", file_content)
    multihost.client[0].run_command("restorecon -v /etc/krb5.keytab")

    def restore():
        """ Restore configuration """
        multihost.master[0].run_command("rm -vf /opt/*")
        multihost.master[0].run_command("cp -vfr /tmp/opt/* /opt")
        multihost.master[0].run_command("rm -vfr /tmp/opt")
        multihost.client[0].run_command("rm -vf /opt/*", raiseonerr=False)
        multihost.client[0].run_command("cp -vfr /tmp/opt/* /opt", raiseonerr=False)
        multihost.client[0].run_command("rm -vfr /tmp/opt", raiseonerr=False)
        multihost.client[0].run_command("cp -vf /etc/krb5.keytab_anuj /etc/krb5.keytab")
        multihost.client[0].run_command("restorecon -v /etc/krb5.keytab")

    request.addfinalizer(restore)


def krb5_fast_setup(client, krb5_use_fast, krb5_fast_principal, **krb5_validate):
    """ To Customize domain parameters for Test Cases
        :client: This will contain client host configuration
        :krb5_use_fast: This will contain value for FAST kerberos pre authentication
                The following options are supported: never/try/demand
        :krb5_fast_principal: This will specify the client principal to use for FAST
                These following options will be tested:
                - Client kerberos principal
                - invalid
                - principal@TEST.TEST
                - null
        :kwargs: This option will append the key value pair (if provided) to sssd.conf, It is expected to get
                key as 'krb5_validate' and 'True' as a value.
    """
    tools = sssdTools(client)
    domain_section = f'domain/{ds_instance_name}'
    domain_params = {
        'krb5_use_fast': krb5_use_fast,
        'krb5_fast_principal': krb5_fast_principal}
    for key, value in krb5_validate.items():
        domain_params[key] = value
    tools.sssd_conf(domain_section, domain_params)
    tools.clear_sssd_cache()


@pytest.fixture(scope='class')
def custom_setup(multihost, setup_sssd_krb, create_posix_usersgroups, krb_connection_timeout):
    """ Added neccessary sssd domain parameters """
    tools = sssdTools(multihost.client[0])
    sssd_params = {'services': "nss, pam"}
    tools.sssd_conf('sssd', sssd_params)
    domain_section = f'domain/{ds_instance_name}'
    domain_params = {'access_provider': 'krb5',
                     'use_fully_qualified_names': 'False',
                     'override_homedir': '/home/%u'}
    tools.sssd_conf(domain_section, domain_params)
    tools.clear_sssd_cache()


@pytest.mark.tier2
@pytest.mark.krbfastprincipal
@pytest.mark.usefixtures('custom_setup', 'create_keytab')
class TestKrbFastPrincipal():
    """
    This is test case class for krb_fast_principal suite

    Test FAST pre authentication by kerberos on the basis of krb5_fast_principal.
    Test for valid and invalid principals, Test for principals with only
    kerberos realm, Test for null principals, Tests for all the previous
    scenarios with krb5_validate set to true.
    """
    @staticmethod
    def test_0001_valid_principal(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_fast_principal: krb5_fast_principal is a valid principal
        :id: d8752706-8ddd-4eaa-a84c-b9b136198c1e
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal in sssd.conf \
             and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal host/$Client@EXAMPLE.TEST in keytab
             Principal matched to the sample (host/$Client@EXAMPLE.TEST)
        """
        krb5_fast_setup(multihost.client[0], 'demand', f'host/{multihost.client[0].sys_hostname}')
        ssh = check_login_client_bool(multihost, "foo1", "Secret123")
        time.sleep(3)
        file = '/var/log/sssd/krb5_child.log'
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert ssh, 'foo1 is not able to login.'
        assert f"Trying to find principal host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST in keytab" \
               in krb5_child_log, f"principal host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST not found in keytab"
        assert f"Principal matched to the sample (host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST)" \
               in krb5_child_log, "Principals did not match"

    @staticmethod
    def test_0002_invalid_principal(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_access_provider: krb5_fast_principal is an invalid principal
        :id: 5c676a70-c094-4936-85b4-e25a5c33d735
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal in sssd.conf \
             and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should not be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal invalid@EXAMPLE.TEST in keytab
             No principal matching invalid@EXAMPLE.TEST found in keytab
        """
        krb5_fast_setup(multihost.client[0], 'try', 'invalid')
        with pytest.raises(Exception):
            check_login_client(multihost, "foo1", "Secret123")
        file = '/var/log/sssd/krb5_child.log'
        time.sleep(3)
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert "Trying to find principal invalid@EXAMPLE.TEST in keytab" in krb5_child_log, \
               "principal invalid@EXAMPLE.TEST not found in keytab"
        assert "No principal matching invalid@EXAMPLE.TEST found in keytab" in krb5_child_log, \
               "principal invalid@EXAMPLE.TEST found in keytab"

    @staticmethod
    def test_0003_principal_at_test_test(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_fast_principal: krb5_fast_principal is principal@TEST.TEST
        :id: 4de7beed-24af-4316-9b8f-88b7deb28e0d
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal in sssd.conf \
             and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should not be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal principal@TEST.TEST in keytab
        """
        krb5_fast_setup(multihost.client[0], 'demand', 'principal@TEST.TEST')
        with pytest.raises(Exception):
            check_login_client(multihost, "foo1", "Secret123")
        time.sleep(3)
        file = '/var/log/sssd/krb5_child.log'
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert "Trying to find principal principal@TEST.TEST in keytab" in krb5_child_log, \
               "principal principal@TEST.TEST not found in keytab"

    @staticmethod
    def test_0004_null_principal(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_fast_principal: krb5_fast_principal is null
        :id: 426d9d4d-4a9b-4de9-b3ee-d1d2bab22c67
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal in sssd.conf \
             and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal (null)@EXAMPLE.TEST in keytab
        """
        krb5_fast_setup(multihost.client[0], 'demand', '')
        ssh = check_login_client_bool(multihost, "foo1", "Secret123")
        time.sleep(3)
        file = '/var/log/sssd/krb5_child.log'
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert ssh, 'foo1 is not able to login.'
        assert "Trying to find principal (null)@EXAMPLE.TEST in keytab" in krb5_child_log, \
               "principal (null)@EXAMPLE.TEST not found in keytab"

    @staticmethod
    def test_0005_valid_principal_and_krb5_validate_true(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_fast_principal: krb5_fast_principal is valid \
            and krb5_validate is true
        :id: 35e8a59a-1c81-4836-b5b8-58f66f6369e1
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal and set krb5_validate to true \
            in sssd.conf and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal host/$Client@EXAMPLE.TEST in keytab
             Principal matched to the sample (host/$Client@EXAMPLE.TEST)
             TGT verified using key for [host/$Client@EXAMPLE.TEST]
        """
        krb5_fast_setup(multihost.client[0], 'demand', f'host/{multihost.client[0].sys_hostname}',
                        krb5_validate='true')
        ssh = check_login_client_bool(multihost, "foo1", "Secret123")
        time.sleep(3)
        file = '/var/log/sssd/krb5_child.log'
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert ssh, 'foo1 is not able to login.'
        assert f"Trying to find principal host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST in keytab"\
            in krb5_child_log, f"principal host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST not found in keytab"
        assert f"Principal matched to the sample (host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST)" \
            in krb5_child_log, "Principals did not match"
        assert f"TGT verified using key for [host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST]" \
            in krb5_child_log, f"TGT did not not verify for [host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST]"

    @staticmethod
    def test_0006_invalid_principal_and_krb5_validate_true(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb5_fast_principal: krb5_fast_principal is invalid \
                and krb5_validate is true
        :id: cdda86ae-2418-427c-a64a-4af277f4b11b
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal and set krb5_validate to true \
            in sssd.conf and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should not be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal invalid@EXAMPLE.TEST in keytab
             No principal matching invalid@EXAMPLE.TEST found in keytab
        """
        krb5_fast_setup(multihost.client[0], 'try', 'invalid', krb5_validate='true')
        with pytest.raises(Exception):
            check_login_client(multihost, "foo1", "Secret123")
        file = '/var/log/sssd/krb5_child.log'
        time.sleep(3)
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert "Trying to find principal invalid@EXAMPLE.TEST in keytab" in krb5_child_log, \
            "principal invalid@EXAMPLE.TEST not found in keytab"
        assert "No principal matching invalid@EXAMPLE.TEST found in keytab" in krb5_child_log, \
            "principal matching found in keytab"

    @staticmethod
    def test_0007_principal_at_test_test_and_krb5_validate_true(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb5_fast_principal: krb5_fast_principal is principal@TEST.TEST \
            and krb5_validate is true
        :id: 73f125d4-8e50-4fe5-a3da-074279220251
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal and set krb5_validate to true \
            in sssd.conf and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should not be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal principal@TEST.TEST in keytab
        """
        krb5_fast_setup(multihost.client[0], 'demand', 'principal@TEST.TEST', krb5_validate='true')
        with pytest.raises(Exception):
            check_login_client(multihost, "foo1", "Secret123")
        time.sleep(3)
        file = '/var/log/sssd/krb5_child.log'
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert "Trying to find principal principal@TEST.TEST in keytab" in krb5_child_log, \
            "principal principal@TEST.TEST not found in keytab"

    @staticmethod
    def test_0008_null_principal_and_krb5_validate_true(multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: krb_provider: krb5_fast_principal: krb5_fast_principal is (null)@EXAMPLE.TEST \
            and krb5_validate is true
        :id: 069691bc-3f07-4eac-ad48-c91c56b9a7ba
        :setup:
          1. Set the values of krb5_use_fast and krb5_fast_principal and set krb5_validate to true \
            in sssd.conf and restart sssd.
        :steps:
          1. Authenticate the user foo1 from the client
          2. Check the krb5_child log for expected messages.
        :expectedresults:
          1. User foo1 should be able to successfully login
          2. Krb5_child Log contains the expected lines:
             Trying to find principal (null)@EXAMPLE.TEST in keytab
        """
        krb5_fast_setup(multihost.client[0], 'demand', '', krb5_validate='true')
        ssh = check_login_client_bool(multihost, "foo1", "Secret123")
        file = '/var/log/sssd/krb5_child.log'
        time.sleep(3)
        krb5_child_log = multihost.client[0].get_file_contents(file).decode('utf-8')
        assert ssh, 'foo1 is not able to login.'
        assert "Trying to find principal (null)@EXAMPLE.TEST in keytab" in krb5_child_log, \
            "principal (null)@EXAMPLE.TEST not found in keytab"
        assert f"TGT verified using key for [host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST]"\
            in krb5_child_log, f"TGT did not not verify for [host/{multihost.client[0].sys_hostname}@EXAMPLE.TEST]"



"""conftest.py for all tests"""

from __future__ import print_function
import subprocess
import random
import os
import posixpath
import pexpect
import pytest
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.ipa.utils import ipaTools
from sssd.testlib.common.utils import ADOperations


pytest_plugins = (
    'sssd.testlib.common.fixtures',
    'pytest_importance',
    'pytest_ticket',
    'sssd.testlib.common.custom_log',
)


def pytest_configure():
    """ Namespace hook to add below dict in the pytest namespace """
    pytest.num_masters = 1
    pytest.num_ad = 1
    pytest.num_atomic = 0
    pytest.num_replicas = 0
    pytest.num_clients = 1
    pytest.num_others = 0

# ====================  Function Scoped Fixtures ==============


@pytest.fixture(scope="function")
def hbac_sshd_rule(session_multihost, request):
    """
    Setup hbac rule for service sshd which allows
    user foobar1 to ssh from client host.
    """
    ipa_server_tools = ipaTools(session_multihost.master[0])
    client_host = session_multihost.client[0].sys_hostname
    ipa_server_tools.add_hbac_rule('test1', 'foobar1', client_host, 'sshd')

    def delete_hbac_rule():
        """ Delete hbac rule """
        ipa_server_tools.del_hbac_rule('test1')
    request.addfinalizer(delete_hbac_rule)


@pytest.fixture(autouse=True)
def capture_sssd_logs(session_multihost, request):
    """This will print sssd logs in case of test failure"""
    yield
    if request.session.testsfailed:
        client = session_multihost.client[0]
        print(f"\n\n===Logs for {request.node.name}===\n\n")
        for data_d in client.run_command("ls /var/log/sssd/").stdout_text.split():
            client.run_command(f'echo "--- {data_d} ---"; '
                               f'cat /var/log/sssd/{data_d}')


@pytest.fixture(scope="function")
def create_aduser_group(session_multihost, request):
    """ create AD user group """
    uid = random.randint(9999, 999999)
    ad = ADOperations(session_multihost.ad[0])
    ad_user = 'testuser%d' % uid
    ad_group = 'testgroup%d' % uid
    ad.create_ad_unix_user_group(ad_user, ad_group)

    def remove_ad_user_group():
        """ Remove windows AD user and group """
        ad.delete_ad_user_group(ad_group)
        ad.delete_ad_user_group(ad_user)

    request.addfinalizer(remove_ad_user_group)
    return ad_user, ad_group


@pytest.fixture(scope="function")
def backup_config_pam_gssapi_services(session_multihost, request):
    """ Take backup of files, Configure domain_params
        Configure /etc/pam.d/sudo
        Configure /etc/pam.d/sudo-i
    """
    tools = sssdTools(session_multihost.client[0])
    domain_name = tools.get_domain_section_name()
    client = sssdTools(session_multihost.client[0])
    domain_params = {'pam_gssapi_services': 'sudo, sudo-i'}
    client.sssd_conf(f'{domain_name}', domain_params)
    session_multihost.client[0].service_sssd('restart')
    session_multihost.client[0].run_command("cp -vf  /etc/pam.d/sudo "
                                            "/etc/pam.d/sudo_bkp")
    session_multihost.client[0].run_command("cp -vf  /etc/pam.d/sudo-i "
                                            "/etc/pam.d/sudo-i_bkp")
    session_multihost.client[0].run_command("sed -i '1 a auth "
                                            "sufficient pam_sss_gss.so' "
                                            "/etc/pam.d/sudo")
    session_multihost.client[0].run_command("sed -i '1 a auth sufficient "
                                            "pam_sss_gss.so' "
                                            "/etc/pam.d/sudo-i")

    def restore():
        session_multihost.client[0].run_command("cp -vf  "
                                                "/etc/pam.d/sudo_bkp "
                                                "/etc/pam.d/sudo")
        session_multihost.client[0].run_command("cp -vf  "
                                                "/etc/pam.d/sudo-i_bkp "
                                                "/etc/pam.d/sudo-i")
        session_multihost.client[0].run_command("rm -vf /tmp/domain_list_*")

    request.addfinalizer(restore)


@pytest.fixture(scope="function")
def create_reverse_zone(session_multihost, request):
    """ Creates reverse zone """
    client_ip = session_multihost.client[0].ip
    subnet = client_ip.split(".", 3)
    del subnet[-1]
    subnet.reverse()
    zone = '.'.join(subnet) + '.in-addr.arpa.'
    policy = 'grant * tcp-self * PTR'

    cmd_createzone = 'ipa dnszone-add %s ' \
                     '--dynamic-update=true ' \
                     '--allow-sync-ptr=true ' \
                     '--skip-overlap-check ' \
                     '--forward-policy=none' % zone
    cmd_modifyzone = 'ipa dnszone-mod %s ' \
                     '--update-policy=\'%s;\'' % (zone, policy)
    session_multihost.master[0].run_command(cmd_createzone,
                                            raiseonerr=False)
    session_multihost.master[0].run_command(cmd_modifyzone,
                                            raiseonerr=False)

    def remove_reverse_zone():
        """  removes reverse zone """
        cmd_removezone = 'ipa dnszone-del %s' % zone
        session_multihost.master[0].run_command(cmd_removezone,
                                                raiseonerr=False)

    request.addfinalizer(remove_reverse_zone)


@pytest.fixture(scope="function")
def default_ipa_groups(session_multihost, request):
    """ Create IPA Groups ipa-group0 to ipa-group9 """
    kinit_admin = 'kinit admin'
    session_multihost.master[0].run_command(kinit_admin,
                                            stdin_text='Secret123',
                                            raiseonerr=False)
    ipa_grp_gid_start = 342156780
    ipa_grp_count = 10
    for i in range(ipa_grp_count):
        gid = ipa_grp_gid_start + i
        group_add = 'ipa group-add ipa-group%d --desc="IPA group%d" --gid %d'\
                    % (i, i, gid)

        cmd = session_multihost.master[0].run_command(group_add,
                                                      raiseonerr=False)
        assert cmd.returncode == 0

    def remove_ipa_groups():
        """ Remove ipa Groups ipa-group0 to ipa-group9 """
        for i in range(10):
            cmd = 'ipa group-del ipa-group%d' % i
            session_multihost.master[0].run_command(cmd)
    request.addfinalizer(remove_ipa_groups)
    return ipa_grp_gid_start


@pytest.fixture(scope="function")
def add_group_member(session_multihost, request):
    """ Add  members in groups of IPA """
    kinit_admin = 'kinit admin'
    session_multihost.master[0].run_command(kinit_admin,
                                            stdin_text='Secret123',
                                            raiseonerr=False)
    for i in range(5):
        for j in range(10):
            add_gr_member = 'ipa group-add-member ipa-group%d ' \
                            '--users=foobar%d' % (i, j)

            cmd = session_multihost.master[0].run_command(add_gr_member,
                                                          raiseonerr=False)
            assert cmd.returncode == 0

    def remove_group_member():
        """ Remove users from IPA groups """
        for i in range(5):
            for j in range(10):
                cmd = ' ipa group-remove-member ipa-group%d --users=foobar%d'\
                      % (i, j)
                session_multihost.master[0].run_command(cmd)
    request.addfinalizer(remove_group_member)


@pytest.fixture(scope='function')
def backupsssdconf(session_multihost, request):
    """ Backup and restore sssd.conf """
    tools = sssdTools(session_multihost.client[0])
    tools.backup_sssd_conf()
    session_multihost.client[0].service_sssd('stop')

    def restoresssdconf():
        """ Restore sssd.conf """
        tools.restore_sssd_conf()
    request.addfinalizer(restoresssdconf)

# ====================  Class Scoped Fixtures ================


@pytest.fixture(scope='class')
def environment_setup(session_multihost, request):
    """
    Install necessary packages
    """
    client = session_multihost.client[0]
    if "Fedora" in client.distro:
        client.run_command("yum install -y shadow-utils*")
    else:
        client.run_command("yum --enablerepo=*-CRB install -y shadow-utils*")

    client.run_command("yum install -y gcc")
    client.run_command("yum install -y podman")
    with pytest.raises(subprocess.CalledProcessError):
        client.run_command("grep subid /etc/nsswitch.conf")
    file_location = "/data/list_subid_ranges.c"
    client.transport.put_file(os.path.dirname(os.path.abspath(__file__))
                              + file_location,
                              '/tmp/list_subid_ranges.c')
    client.run_command("gcc /tmp/list_subid_ranges.c"
                       "  -lsubid -o /tmp/list_subid_ranges")

    def remove():
        """ Remove file """
        for file in ['list_subid_ranges', 'list_subid_ranges.c']:
            client.run_command(f"rm -vf /tmp/{file}")

    request.addfinalizer(remove)


@pytest.fixture(scope='class')
def subid_generate(session_multihost, request):
    """
    Generate subid for user admin
    """
    user = "admin"
    test_password = "Secret123"

    p_ssh = pexpect.pxssh.pxssh(
        options={"StrictHostKeyChecking": "no",
                 "UserKnownHostsFile": "/dev/null"}
    )
    p_ssh.force_password = True
    try:
        p_ssh.login(session_multihost.client[0].ip, user, test_password)
        p_ssh.sendline('kinit')
        p_ssh.expect('Password for .*:', timeout=10)
        p_ssh.sendline(test_password)
        p_ssh.prompt(timeout=5)
        p_ssh.sendline(f'ipa subid-generate --owner={user}; echo "retcode:$?"')
        p_ssh.prompt(timeout=5)
        subid_gen_out = str(p_ssh.before)
        p_ssh.logout()
    except pexpect.pxssh.ExceptionPxssh:
        pytest.fail("Failed to login via ssh.")
    assert "retcode:0" in subid_gen_out, "Generating subid range failed."


@pytest.fixture(scope='class')
def bkp_cnfig_for_subid_files(session_multihost, request):
    """ Back up files used in test
        And config /etc/nsswitch.conf
    """
    session_multihost.client[0].run_command("cp -vf  "
                                            "/etc/subuid "
                                            "/tmp/subuid_bkp")
    session_multihost.client[0].run_command("cp -vf  "
                                            "/etc/subgid "
                                            "/tmp/subgid_bkp")
    session_multihost.client[0].run_command("cp -vf  "
                                            "/etc/nsswitch.conf "
                                            "/tmp/nsswitch.conf_bkp")
    session_multihost.client[0].run_command("echo 'subid: sss'  "
                                            ">> /etc/nsswitch.conf")

    def restore():
        """ Restore """
        session_multihost.client[0].run_command("mv -vf  "
                                                "/tmp/subuid_bkp "
                                                "/etc/subuid")
        session_multihost.client[0].run_command("mv -vf  "
                                                "/tmp/subgid_bkp "
                                                "/etc/subgid")
        session_multihost.client[0].run_command("mv -vf  "
                                                "/tmp/nsswitch.conf_bkp "
                                                "/etc/nsswitch.conf")
    request.addfinalizer(restore)


@pytest.fixture(scope="class")
def default_ipa_users(session_multihost, request):
    """ Create IPA Users foobar0 to foobar9 """
    kinit_admin = 'kinit admin'
    session_multihost.master[0].run_command(kinit_admin,
                                            stdin_text='Secret123',
                                            raiseonerr=False)
    for i in range(10):
        user_info = {'firstname': 'Foo',
                     'lastname': 'bar%d' % i,
                     'loginname': 'foobar%d' % i,
                     'default_password': 'RedHat@123',
                     'reset_password': 'Secret123'}
        useradd = "echo '%s' | ipa user-add --first %s " \
                  " --last %s --password %s" % (user_info['default_password'],
                                                user_info['firstname'],
                                                user_info['lastname'],
                                                user_info['loginname'])
        cmd = session_multihost.master[0].run_command(useradd,
                                                      raiseonerr=False)
        assert cmd.returncode == 0

    def remove_ipa_users():
        """ Remove ipa users foobar1 to foobar10 """
        for i in range(10):
            cmd = 'ipa user-del foobar%d' % i
            session_multihost.master[0].run_command(cmd)

    request.addfinalizer(remove_ipa_users)


@pytest.fixture(scope="class")
def reset_password(session_multihost, request):
    """ Reset passwords for users foobar0 to foobar9 """
    for i in range(10):
        user = 'foobar%d' % i
        cmd = "echo -e 'RedHat@123\nSecret123\nSecret123' | kinit %s" % user
        session_multihost.client[0].run_command(cmd)
        kdestroy = 'kdestroy'
        session_multihost.client[0].run_command(kdestroy)


@pytest.fixture(scope="class")
def disable_allow_all_hbac(session_multihost, request):
    """ Disable allow_all hbac rule """
    disable_allow_all = 'ipa hbacrule-disable allow_all'
    try:
        session_multihost.master[0].run_command(disable_allow_all)
    except subprocess.CalledProcessError:
        pytest.fail("Failed to disable allow_all rule")

    def allow_all_hbac():
        """ Enable allow_all hbac rule """
        allow_all = 'ipa hbacrule-enable allow_all'
        try:
            session_multihost.master[0].run_command(allow_all)
        except subprocess.CalledProcessError:
            pytest.fail("Failed to enable allow_all rule")

    request.addfinalizer(allow_all_hbac)


@pytest.fixture(scope="class")
def multihost(session_multihost, request):
    """ Multihost fixture to be used by tests
    :param obj session_multihost: multihost object
    :return obj session_multihost: return multihost object
    """
    if hasattr(request.cls(), 'class_setup'):
        request.cls().class_setup(session_multihost)
        request.addfinalizer(
            lambda: request.cls().class_teardown(session_multihost))
    return session_multihost


@pytest.fixture(scope="class")
def create_ad_users(session_multihost, request):
    """ Create AD Users """
    cwd = os.path.dirname(os.path.abspath(__file__))
    file_list = ['users.csv', 'add-users.ps1', 'remove-users.ps1']
    for filename in file_list:
        remote_file_path = posixpath.join('/home/administrator', filename)
        source_file_path = posixpath.join(cwd, filename)
        session_multihost.ad[0].transport.put_file(source_file_path,
                                                   remote_file_path)
    user_add_cmd = 'powershell -inputformat none -noprofile ./add-users.ps1'
    session_multihost.ad[0].run_command(user_add_cmd, raiseonerr=False)

    def remove_users():
        """ Remove AD users """
        del_cmd = 'powershell -inputformat none -noprofile ./remove-users.ps1'
        session_multihost.ad[0].run_command(del_cmd, raiseonerr=False)

    request.addfinalizer(remove_users)


@pytest.fixture(scope="class")
def create_ad_groups(session_multihost, request):
    """ Create AD Groups """
    cwd = os.path.dirname(os.path.abspath(__file__))
    file_list = ['groups.csv', 'nestedgroups.csv',
                 'add-groups.ps1', 'remove-groups.ps1']
    for filename in file_list:
        remote_file_path = posixpath.join('/home/administrator', filename)
        source_file_path = posixpath.join(cwd, filename)
        session_multihost.ad[0].transport.put_file(source_file_path,
                                                   remote_file_path)
    add_cmd = 'powershell -inputformat none -noprofile ./add-groups.ps1'
    session_multihost.ad[0].run_command(add_cmd, raiseonerr=False)

    def remove_ad_groups():
        """ Remove AD Groups """
        del_cmd = 'powershell -inputformat none -noprofile ./remove-groups.ps1'
        session_multihost.ad[0].run_command(del_cmd, raiseonerr=False)

    request.addfinalizer(remove_ad_groups)


# ====================  Session Scoped Fixtures ================

@pytest.fixture(scope="session", autouse=True)
# pylint: disable=unused-argument
def setup_ipa_client(session_multihost, request):
    """ Setup ipa client """
    sssd_client = sssdTools(session_multihost.client[0])
    client_hostname = session_multihost.client[0].sys_hostname
    server_hostname = session_multihost.master[0].sys_hostname
    ipa_client = ipaTools(session_multihost.client[0])
    ipa_server = ipaTools(session_multihost.master[0])
    client = sssdTools(session_multihost.client[0])
    client.client_install_pkgs()
    ipa_client.install_common_pkgs()
    ipa_server.install_common_pkgs()
    ipa_client_ip = session_multihost.client[0].ip
    ipa_server_uuid = ipa_server.get_default_nw_uuid()
    ipa_server_ip = ipa_server.get_interface_ip(ipa_server_uuid)
    sssd_client.update_resolv_conf(ipa_server_ip)
    options = "--ip-address=%s --hostname %s "\
              "--server %s --domain %s "\
              "--realm %s -w %s -p %s -U --mkhomedir" % (ipa_client_ip,
                                                         client_hostname,
                                                         server_hostname,
                                                         "testrealm.test",
                                                         "TESTREALM.TEST",
                                                         "Secret123",
                                                         "admin")
    client_install_cmd = "ipa-client-install %s" % options
    try:
        session_multihost.client[0].run_command(client_install_cmd)
    except subprocess.CalledProcessError:
        pytest.fail("ipa client install failed")

    def teardown_session():
        """ Uninstall ipa client from server """
        client_uninstall_cmd = 'ipa-client-install --uninstall -U'
        session_multihost.client[0].run_command(client_uninstall_cmd)
    request.addfinalizer(teardown_session)



""" Automation of Netgroup suite

:requirement: netgroup
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

import time
import ldap
import pytest
from sssd.testlib.common.utils import sssdTools, LdapOperations
from constants import ds_instance_name, ds_suffix


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups', 'netgroups')
@pytest.mark.netgroup
class TestNetgroup(object):

    @pytest.mark.tier1
    def test_0001_bz1502686(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: netgroup: SSSD crashes in nss
         responder after netgroup timeout when backend is offline
        :id: 2e5823a2-835b-45fd-a2be-36b20856e726
        :customerscenario: True
        """
        hostname = multihost.master[0].sys_hostname
        bad_ldap_uri = "ldaps://typo.%s" % hostname
        # stop sssd service
        multihost.client[0].service_sssd('stop')
        tools = sssdTools(multihost.client[0])
        bkupconf = 'cp -f /etc/sssd/sssd.conf /etc/sssd/sssd.conf.bkup'
        restoreconf = 'cp -f /etc/sssd/sssd.conf.bkup /etc/sssd/sssd.conf'
        remove_bkup = 'rm -f /etc/sssd/sssd.conf.bkup'
        multihost.client[0].run_command(bkupconf)
        # remove sssd cache
        tools.remove_sss_cache('/var/lib/sss/db')
        domain_params = {'ldap_uri': bad_ldap_uri}
        tools.sssd_conf('domain/%s' % ds_instance_name, domain_params)
        start = multihost.client[0].service_sssd('start')
        # Check backend status
        domain_status = "sssctl domain-status %s -o" % ds_instance_name
        chk_status = multihost.client[0].run_command(domain_status,
                                                     raiseonerr=False)
        if 'Offline' in chk_status.stdout_text.strip():
            status = 'PASS'
        pid_nss = "pidof sssd_nss"
        chk_pid1 = multihost.client[0].run_command(pid_nss, raiseonerr=False)
        pid_nss1 = chk_pid1.stdout_text.strip()
        if start == 0:
            # request for netgroup
            getent = 'getent netgroup -s sss netgroup_1'
            chk_req = multihost.client[0].run_command(getent, raiseonerr=False)
            if chk_req.returncode != 2:
                status = 'FAIL'
            time.sleep(16)
            sssd_proc = ['sssd', 'sssd_be', 'sssd_nss', 'sssd_pam']
            for proc in sssd_proc:
                pgrep = 'pgrep %s' % proc
                cmd = multihost.client[0].run_command(pgrep, raiseonerr=False)
                if cmd.returncode != 0:
                    status = 'FAIL'
                    print("%s process failed to start" % proc)
                else:
                    pid2 = multihost.client[0].run_command(pid_nss)
                    # automation of bz1576852
                    pid_nss2 = pid2.stdout_text.strip()
                    if pid_nss1 != pid_nss2:
                        status = 'FAIL'
                        print("nss process crashed and restarted")
                    else:
                        status = 'PASS'
            multihost.client[0].run_command(restoreconf)
            multihost.client[0].run_command(remove_bkup)
            assert status != 'FAIL'

    @pytest.mark.tier1
    def test_0002_bz1406437(self, multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: netgroup: sssctl netgroup-show
         Cannot allocate memory
        :id: bdcf03d6-fe5d-44fd-9517-f812dacaf6af
        """
        multihost.client[0].service_sssd('stop')
        tools = sssdTools(multihost.client[0])
        tools.remove_sss_cache('/var/lib/sss/db')
        multihost.client[0].service_sssd('start')
        time.sleep(10)
        getent = 'getent netgroup netgroup_1@%s' % ds_instance_name
        cmd = multihost.client[0].run_command(getent, raiseonerr=False)
        assert cmd.returncode == 0
        sssctl = 'sssctl netgroup-show netgroup_1@%s' % ds_instance_name
        cmd = multihost.client[0].run_command(sssctl, raiseonerr=False)
        if cmd.returncode == 0:
            search = 'Cannot allocate memory'
            out = cmd.stdout_text
            assert out.find(search) == -1
        else:
            pytest.fail("%s failed" % sssctl)

    @pytest.mark.tier1
    def test_0003_background_refresh(self, multihost):
        """
        :title: netgroup: background refresh task does not refresh
         updated netgroup entries
        :id: b17d904d-0d64-4f4a-bbad-4c7f63e1faf2
        :bugzilla:
         https://bugzilla.redhat.com/show_bug.cgi?id=1779486 (RHEL8.2)
         https://bugzilla.redhat.com/show_bug.cgi?id=1822461 (RHEL7.8)
        """
        multihost.client[0].service_sssd('stop')
        tools = sssdTools(multihost.client[0])
        tools.remove_sss_cache('/var/lib/sss/db')
        domain_params = {'entry_cache_timeout': '30',
                         'refresh_expired_interval': '22'}
        tools.sssd_conf('domain/%s' % ds_instance_name, domain_params)
        multihost.client[0].service_sssd('restart')
        # getent netgroup_1
        getent_cmd = "getent netgroup netgroup_1"
        multihost.client[0].run_command(getent_cmd)
        shortname = multihost.client[0].sys_hostname.strip().split('.')[0]
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        netgroup_dn = 'cn=netgroup_1,ou=Netgroups,%s' % (ds_suffix)
        nisNetgroupTriple = "(%s,foo1,%s)" % (shortname, ds_suffix)
        modify_netgroup = [(ldap.MOD_REPLACE, 'nisNetgroupTriple',
                            nisNetgroupTriple.encode('utf-8'))]
        (_, _) = ldap_inst.modify_ldap(netgroup_dn, modify_netgroup)
        time.sleep(40)
        ldb_cmd = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb'\
                  ' -b cn=Netgroups,cn=%s,cn=sysdb' % (ds_instance_name,
                                                       ds_instance_name)
        cmd = multihost.client[0].run_command(ldb_cmd)
        new_entry = "netgroupTriple: (%s,foo1,%s)" % (shortname, ds_suffix)
        tools.sssd_conf('domain/%s' % ds_instance_name,
                        domain_params, action='delete')
        assert new_entry in cmd.stdout_text.strip().split('\n')



""" CIFS Test cases

:requirement: cifs
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import time
import pytest


@pytest.fixture(scope="class", autouse=True)
def add_etc_host_records(session_multihost, request):
    """ Add master and AD to /etc/hosts so they can be resolved
    :param obj session_multihost: multihost object
    :param obj request: pytest request object
    """
    session_multihost.client[0].run_command(
        'cp /etc/hosts /etc/hosts.orig', raiseonerr=False)
    records = f"{session_multihost.master[0].ip} " \
        f"{session_multihost.master[0].sys_hostname.strip().split('.')[0]} " \
        f"{session_multihost.master[0].sys_hostname.strip()}\\n" \
        f"{session_multihost.ad[0].ip} " \
        f"{session_multihost.ad[0].sys_hostname.strip().split('.')[0]} " \
        f"{session_multihost.ad[0].sys_hostname.strip()}\\n"
    session_multihost.client[0].run_command(
        f'echo -e "\\n{records}" >> /etc/hosts', raiseonerr=False)
    session_multihost.client[0].run_command('cat /etc/hosts', raiseonerr=False)

    def restore_etc_hosts():
        """ Restore backed up hosts"""
        session_multihost.client[0].run_command(
            'cp -f /etc/hosts.orig /etc/hosts', raiseonerr=False)
    request.addfinalizer(restore_etc_hosts)


@pytest.mark.usefixtures('winbind_server',
                         'configure_samba',
                         'samba_share_permissions')
@pytest.mark.tier2
@pytest.mark.cifs
class Testcifs(object):
    """ Samba IDMAP and CIFS Automations

    :setup:
      1. Join RHEL system to windows AD domain using below command
         $ realm join -v TESTRELM.TEST --membership-software=samba
      2. configure smb.conf to use sss as idmap backend
         idmap config * : backend = sss
         idmap config * : range   = 200000-2147483647
      3. Restart winbind
    """

    @pytest.mark.tier1
    def test_0001_wbinfo(self, multihost):
        """
        :title: IDM-SSSD-TC: samba_idmap: Samba can not
         register sss idmap module due to SMB_IDMAP_INTERFACE_VERSION
        :id: bad2770e-75de-4b41-af47-e9a38b8c0e73
        :requirement: IDM-SSSD-REQ: Samba with sssd as idmap backend
        :steps: Run wbinfo -i <DOMAIN>\\administrator
        :expectedresults: wbinfo command should be successfull
        """
        realm = multihost.ad[0].realm
        wb_cmd = 'wbinfo -i {}{}{}'.format(realm, '\\\\', "administrator")
        cmd = multihost.client[0].run_command(wb_cmd, raiseonerr=False)
        assert cmd.returncode == 0

    @pytest.mark.tier2
    def test_0002_smb1mount(self, multihost):
        """
        :title: cifs: mount samba share using smb1 protocol
        :id: b82b6c93-1857-449c-bfbb-2a184e12cce6
        """
        ad_user = 'idmfoouser1'
        kinit = 'kinit %s' % ad_user
        server = multihost.master[0].sys_hostname.strip().split('.')[0]
        multihost.client[0].run_command(kinit, stdin_text='Secret123')
        mountcifs = "mount -t cifs -o cifsacl "\
                    "-o sec=krb5 -o username=%s //%s/share1"\
                    " /mnt/samba/share1" % (ad_user, server)
        cmd = multihost.client[0].run_command(mountcifs, raiseonerr=False)
        if cmd.returncode != 0:
            journalctl = 'journalctl -x -n 50 --no-pager'
            multihost.client[0].run_command(journalctl)
        else:
            assert cmd.returncode == 0
        umount = "umount /mnt/samba/share1"
        multihost.client[0].run_command(umount, raiseonerr=False)
        kdestroy = 'kdestroy -A'
        multihost.client[0].run_command(kdestroy, raiseonerr=False)

    def test_0003_smb3mount(self, multihost):
        """
        :title: cifs: mount samba share using encrypted smb3 protocol
        :id: 1b10f6f2-d91e-4ea4-a56a-985cd9f9d884
        """
        ad_user = 'idmfoouser1'
        ad_group = 'idmfoogroup1'
        kinit = 'kinit %s' % ad_user
        server = multihost.master[0].sys_hostname.strip().split('.')[0]
        share_path = '/mnt/samba/share1'
        multihost.client[0].run_command(kinit, stdin_text='Secret123')
        mountcifs = "mount -t cifs -o username=%s,seal,sec=krb5,vers=3.0"\
                    " //%s/share1 /mnt/samba/share1" % (ad_user, server)
        cmd = multihost.client[0].run_command(mountcifs, raiseonerr=False)
        if cmd.returncode != 0:
            journalctl = 'journalctl -x -n 50 --no-pager'
            multihost.client[0].run_command(journalctl)
            pytest.fail("mounting using smb3 protocol failed")
        perms = '660'
        echo1 = "echo 'testwrite1' > "\
                "%s/allgroup/foobar.%s.%s.txt" % (share_path,
                                                  perms, ad_user)
        echo2 = "echo 'testwrite1' > "\
                "%s/%s/foobar.%s.%s.txt" % (share_path,
                                            ad_group, perms, ad_user)
        cmd = multihost.client[0].run_command(echo1, raiseonerr=False)
        assert cmd.returncode == 0
        multihost.client[0].run_command(echo2, raiseonerr=False)
        assert cmd.returncode == 0
        remove_file = "rm -f %s/allgroup/foobar.%s.%s.txt" % (share_path,
                                                              perms, ad_user)
        multihost.client[0].run_command(remove_file)
        remove_file = "rm -f %s/%s/foobar.%s.%s.txt" % (share_path,
                                                        ad_group,
                                                        perms, ad_user)
        multihost.client[0].run_command(remove_file)
        umount = "umount /mnt/samba/share1"
        multihost.client[0].run_command(umount, raiseonerr=False)
        kdestroy = 'kdestroy -A'
        multihost.client[0].run_command(kdestroy, raiseonerr=False)

    def test_0004_writeable(self, multihost):
        """
        :title: cifs: verify samba share is writeable
        :id: abd21985-ebd3-4c9d-987c-6e6da32bb922
        """
        realm = multihost.ad[0].realm
        for idx in range(1, 3):
            ad_user = 'idmfoouser%d' % idx
            ad_group = 'idmfoogroup%d' % idx
            kinit = 'kinit %s' % ad_user
            server = multihost.master[0].sys_hostname.strip().split('.')[0]
            share_path = '/mnt/samba/share1'
            multihost.client[0].run_command(kinit, stdin_text='Secret123')
            mountcifs = "mount -t cifs -o cifsacl "\
                        "-o sec=krb5 -o username=%s //%s/share1"\
                        " /mnt/samba/share1" % (ad_user, server)
            cmd = multihost.client[0].run_command(mountcifs, raiseonerr=False)
            assert cmd.returncode == 0
            time.sleep(5)
            journalctl = 'journalctl -x -n 50 --no-pager'
            multihost.client[0].run_command(journalctl)
            getent = 'getent passwd %s@%s' % (ad_user, realm)
            file_perms = ['660', '640', '600', '660']
            multihost.client[0].run_command(getent)
            for perms in file_perms:
                all_file = "%s/allgroup/testfile.%s.%s.txt" % (share_path,
                                                               perms, ad_user)
                group_file = "%s/%s/testfile.%s.%s.txt" % (share_path,
                                                           ad_group,
                                                           perms, ad_user)
                echo1 = "echo 'testwrite1' > %s" % (all_file)
                echo2 = "echo 'testwrite1' > %s" % (group_file)
                cmd = multihost.client[0].run_command(echo1, raiseonerr=False)
                assert cmd.returncode == 0
                cmd = multihost.client[0].run_command(echo2, raiseonerr=False)
                assert cmd.returncode == 0
                remove_file = "rm -f %s " % (all_file)
                multihost.client[0].run_command(remove_file)
                remove_file = "rm -f %s" % (group_file)
                multihost.client[0].run_command(remove_file)
            umount = "umount /mnt/samba/share1"
            multihost.client[0].run_command(umount, raiseonerr=False)
            kdestroy = 'kdestroy -A'
            multihost.client[0].run_command(kdestroy, raiseonerr=False)

    def test_0005_aclcheck(self, multihost, cifsmount):
        """
        :title: cifs: verify cifs acls on samba share with smb1 mount
        :id: b6f86e6b-03f7-4d76-8450-9b2d3d1ed71e
        """
        ad_user = 'idmfoouser1'
        ad_group = 'idmfoogroup1'
        netbiosname = multihost.ad[0].netbiosname.strip()
        share_path = '/mnt/samba/share1'
        wb_cmd = 'wbinfo -n {}{}{}'.format(netbiosname, '\\\\', ad_user)
        cmd = multihost.client[0].run_command(wb_cmd, raiseonerr=False)
        smbuser_sid = cmd.stdout_text.split(' ')[0]
        wb_cmd = 'wbinfo -n {}{}{}'.format(netbiosname, '\\\\', ad_group)
        cmd = multihost.client[0].run_command(wb_cmd, raiseonerr=False)
        smbpgroup_sid = cmd.stdout_text.split(' ')[0]
        wb_cmd = 'wbinfo -n {}{}{}'.format(netbiosname, '\\\\',
                                           "idmfooallgroup")
        cmd = multihost.client[0].run_command(wb_cmd, raiseonerr=False)
        smballgroup_sid = cmd.stdout_text.split(' ')[0]
        file_perms = ['660', '640', '600', '660']
        for perms in file_perms:
            allgroup_file = "%s/allgroup/testfoo.%s.%s.txt" % (share_path,
                                                               perms, ad_user)
            acl1 = "getcifsacl -r %s" % allgroup_file
            adgroup_file = "%s/%s/testfoo.%s.%s.txt" % (share_path, ad_group,
                                                        perms, ad_user)
            acl2 = "getcifsacl -r %s" % adgroup_file
            echo1 = "echo 'testwrite1' > %s" % (allgroup_file)

            echo2 = "echo 'testwrite1' > %s" % (adgroup_file)
            cmd = multihost.client[0].run_command(echo1, raiseonerr=False)
            assert cmd.returncode == 0
            getcifsacl = multihost.client[0].run_command(acl1,
                                                         raiseonerr=False)
            usersid = "ACL:%s:0x0/0x0/0x1e01ff" % smbuser_sid
            allgroupsid = "ACL:%s:0x0/0x0/0x120089" % smballgroup_sid
            usergroupsid = "ACL:%s:0x0/0x0/0x120089" % smbpgroup_sid
            acl_list = getcifsacl.stdout_text.split('\n')
            assert usersid in acl_list
            assert allgroupsid in acl_list
            cmd = multihost.client[0].run_command(echo2, raiseonerr=False)
            assert cmd.returncode == 0
            getcifsacl = multihost.client[0].run_command(acl2,
                                                         raiseonerr=False)
            acl_list = getcifsacl.stdout_text.split('\n')
            assert usergroupsid in acl_list
            assert usersid in acl_list

    def test_0006_readfromclient(self, multihost, cifsmount):
        """
        :title: verify files modified on server
         are reflected properly on client
        :id: d8861167-2815-4aaa-abeb-1ac9183e3627
        """
        ad_user = 'idmfoouser1'
        ad_group = 'idmfoogroup1'
        share_path = '/mnt/samba/share1'
        file_perms = ['660', '640', '600', '660']
        for perms in file_perms:
            allgroup_file = "%s/allgroup/testfoo1.%s.%s.txt" % (share_path,
                                                                perms, ad_user)
            adgroup_file = "%s/%s/testfoo1.%s.%s.txt" % (share_path, ad_group,
                                                         perms, ad_user)
            echo1 = "echo 'testwrite1' > %s" % (allgroup_file)
            echo2 = "echo 'testwrite1' > %s" % (adgroup_file)
            cmd = multihost.master[0].run_command(echo1, raiseonerr=False)
            assert cmd.returncode == 0
            cmd = multihost.master[0].run_command(echo2, raiseonerr=False)
            assert cmd.returncode == 0
            readfile1 = "cat %s" % (allgroup_file)
            cmd = multihost.client[0].run_command(readfile1, raiseonerr=False)
            assert 'testwrite1' in cmd.stdout_text
            readfile2 = "cat %s" % (adgroup_file)
            cmd = multihost.client[0].run_command(readfile2, raiseonerr=False)
            assert 'testwrite1' in cmd.stdout_text

    def test_0007_readfromserver(self, multihost, cifsmount):
        """
        :title: verify files modified on client
         are reflected properly on server
        :id: 0b8e677b-159b-49c0-9b85-3effdd22642d
        """
        ad_user = 'idmfoouser1'
        ad_group = 'idmfoogroup1'
        share_path = '/mnt/samba/share1'
        file_perms = ['660', '640', '600', '660']
        for perms in file_perms:
            allgroup_file = "%s/allgroup/testfoo2.%s.%s.txt" % (share_path,
                                                                perms, ad_user)
            adgroup_file = "%s/%s/testfoo2.%s.%s.txt" % (share_path, ad_group,
                                                         perms, ad_user)
            echo1 = "echo 'testwrite1' > %s" % (allgroup_file)
            echo2 = "echo 'testwrite1' > %s" % (adgroup_file)
            cmd = multihost.client[0].run_command(echo1, raiseonerr=False)
            assert cmd.returncode == 0
            cmd = multihost.client[0].run_command(echo2, raiseonerr=False)
            assert cmd.returncode == 0
            readfile1 = "cat %s" % (allgroup_file)
            cmd = multihost.master[0].run_command(readfile1, raiseonerr=False)
            assert 'testwrite1' in cmd.stdout_text
            readfile2 = "cat %s" % (adgroup_file)
            cmd = multihost.master[0].run_command(readfile2, raiseonerr=False)
            assert 'testwrite1' in cmd.stdout_text



"""
Proxy Provider tests.

:requirement: Proxy Provider
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology


@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
def test_proxy__lookup_and_authenticate_user_using_pam_ldap_and_nslcd(client: Client, ldap: LDAP):
    """
    :title: Lookup and authenticate user using PAM LDAP and NSLCD.
    :setup:
        1. Setup SSSD to use PAM LDAP and NSLCD.
        2. Create OU, and create a user in the new OU.
    :steps:
        1. Lookup user.
        2. Login in as user.
    :expectedresults:
        1. User found.
        2. User logged in.
    :customerscenario: True
    """
    client.sssd.common.proxy("ldap", ["id", "auth", "chpass"], server_hostname=ldap.host.hostname)
    client.sssd.svc.restart("nslcd")
    client.sssd.restart()
    ou_users = ldap.ou("users").add()
    user = ldap.user("user-1", basedn=ou_users).add(uid=10001, gid=10001, password="Secret123")

    assert client.tools.id(user.name) is not None, "User not found!"
    assert client.auth.ssh.password(user.name, password="Secret123"), "User login failed!"


@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.ticket(bz=895570)
def test_proxy__lookup_user_using_pam_ldap_and_nslcd_with_proxy_fast_alias_enabled(client: Client, ldap: LDAP):
    """
    :title: Lookup user using PAM LDAP and NSLCD with proxy_fast_alias enabled.
    :description: This bugzilla was created to squash 'ldb_modify failed' message when proxy_fast_alias is enabled.
    :setup:
        1. Setup SSSD to use PAM LDAP and NSLCD and set "proxy_fast_alias = true".
        2. Create OU, and create a user in the new OU.
    :steps:
        1. Lookup user.
        2. Check logs for ldb_modify errors.
    :expectedresults:
        1. User found.
        2. No error messages in log.
    :customerscenario: True
    """
    client.sssd.common.proxy("ldap", ["id", "auth", "chpass"], server_hostname=ldap.host.hostname)
    client.sssd.domain["proxy_fast_alias"] = "True"
    client.sssd.svc.restart("nslcd")
    client.sssd.restart()
    ou_users = ldap.ou("users").add()
    user = ldap.user("user-1", basedn=ou_users).add(uid=10001, gid=10001, password="Secret123")

    assert client.tools.id(user.name) is not None, "User not found!"

    log = client.fs.read(client.sssd.logs.domain())
    assert "ldb_modify failed: [Invalid attribute syntax]" not in log, "'ldb_modify failed' message found in logs!"



""" Automation of configuration merging

:requirement: IDM-SSSD-REQ: Configuration merging
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import re
from random import choice
from string import ascii_uppercase
from constants import ds_instance_name
from sssd.testlib.common.utils import sssdTools
import pytest


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.configmerge
class TestConfigMerge(object):
    """
    This is test case class for ldap configuration merging suite
    """
    @pytest.mark.tier1
    def test_0001_verifypermission(self, multihost):
        """
        :title: IDM-SSSD-TC: Configuration merging: Verify the permission of
         snippet files
        :id: 6e52a34b-0dda-4c84-8964-72915026ec8c
       """
        multihost.client[0].service_sssd('stop')
        section = "domain/%s" % ds_instance_name
        content = "[%s]\nuse_fully_quailified_name = False" % section
        snippet_file = "/etc/sssd/conf.d/01_snippet.conf"
        multihost.client[0].put_file_contents(snippet_file, content)
        returncode = multihost.client[0].service_sssd('start')
        if returncode == 0:
            config_check = 'sssctl config-check'
            cmd = multihost.client[0].run_command(config_check,
                                                  raiseonerr=False)
            log_1 = re.compile(r'.*did\snot\spass\saccess\scheck.*')
            if log_1.search(cmd.stdout_text):
                user = 'foo1@%s' % ds_instance_name
                cmd = 'getent passwd %s' % user
                lookup = multihost.client[0].run_command(cmd, raiseonerr=False)
                assert lookup.returncode == 0
                user = 'foo1'
                cmd = 'getent passwd %s' % user
                lookup = multihost.client[0].run_command(cmd, raiseonerr=False)
                assert lookup.returncode == 2
        cmd = 'rm -f %s' % snippet_file
        multihost.client[0].run_command(cmd, raiseonerr=False)

    @pytest.mark.tier1
    def test_0002_hiddenfiles(self, multihost):
        """
        :title: IDM-SSSD-TC: Configuration merging: SSSD reads all *.conf
         files, that are not starting with a . (hidden files)
        :id: 2428bb6b-535d-46b9-a092-a9c9c3f141fa
        """
        multihost.client[0].service_sssd('stop')
        dom_section = "domain/%s" % ds_instance_name
        file_content = "[%s]\nuse_fully_quailified_name = False" % dom_section
        snippet_file = "/etc/sssd/conf.d/._01_snippet.conf"
        multihost.client[0].put_file_contents(snippet_file, file_content)
        start = multihost.client[0].service_sssd('start')
        sssctl_cmd = 'sssctl config-check'
        if start == 0:
            cmd = multihost.client[0].run_command(sssctl_cmd, raiseonerr=False)
            assert cmd.returncode == 0
        remove_snippet = 'rm -f /etc/sssd/conf.d/._01_snippet.conf'
        multihost.client[0].run_command(remove_snippet, raiseonerr=False)

    @pytest.mark.tier1
    def test_0003_lastreadparameter(self, multihost):
        """
        :title: IDM-SSSD-TC: Configuration merging: SSSD will use the last
         read parameter if the same option appears multiple times.
        :id: 57b4f8c5-08fe-46af-b7e3-e8271d72e41e
        """
        multihost.client[0].service_sssd('stop')
        dom_section = "domain/%s" % ds_instance_name
        for idx in ['True', 'False']:
            content = "[%s]\nuse_fully_qualified_names = %s" % (dom_section,
                                                                idx)
            snippet_file = "/etc/sssd/conf.d/%s_snippet.conf" % idx
            multihost.client[0].put_file_contents(snippet_file, content)
            chmod = 'chmod 600 %s' % snippet_file
            multihost.client[0].run_command(chmod)
        start = multihost.client[0].service_sssd('start')
        cmd = multihost.client[0].run_command(['sssctl', 'config-check'])
        if cmd.returncode == 0 and start == 0:
            for user in ['foo1@%s' % ds_instance_name, 'foo1']:
                getent = 'getent passwd %s' % user
                cmd = multihost.client[0].run_command(getent, raiseonerr=False)
                if '@' not in getent:
                    assert cmd.returncode != 0
                else:
                    assert cmd.returncode == 0
        for str1 in ['True', 'False']:
            cmd = 'rm -f /etc/sssd/conf.d/%s_snippet.conf' % str1
            multihost.client[0].run_command(cmd, raiseonerr=False)

    @pytest.mark.tier1
    def test_0004_formatsnippetfile(self, multihost):
        """
        :title: IDM-SSSD-TC: Configuration merging: Verify the format of
         snippet files
        :id: 8496df1b-32ef-45cd-9d9f-34dcd18656ad
        """
        multihost.client[0].service_sssd('stop')
        dom_section = "domain/%s" % ds_instance_name
        file_content = "[%s]\nuse_fully_qualified_names = False" % dom_section
        snippet_file = "/etc/sssd/conf.d/01_snippet.conf.disable"
        multihost.client[0].put_file_contents(snippet_file, file_content)
        cmd_chmod = 'chmod 600 %s' % snippet_file
        multihost.client[0].run_command(cmd_chmod, raiseonerr=False)
        start = multihost.client[0].service_sssd('start')
        if start == 0:
            config_check = 'sssctl config-check'
            cmd = multihost.client[0].run_command(config_check,
                                                  raiseonerr=False)
            assert cmd.returncode == 0
        else:
            assert False
        remove_snippet = 'rm -f %s' % snippet_file
        multihost.client[0].run_command(remove_snippet, raiseonerr=False)

    @pytest.mark.tier1
    def test_0005_ownershisnippetfile(self, multihost):
        """
        :title: IDM-SSSD-TC: Configuration merging: Verify the ownership of
         snippet files
        :id: b89f0403-67ee-4f6b-8c17-bdc4f5213d4f
        """
        multihost.client[0].service_sssd('stop')
        gen = sssdTools(multihost.client[0])
        gen.remove_sss_cache('/var/lib/sss/db')
        dom_section = "domain/%s" % ds_instance_name
        file_content = "[%s]\nuse_fully_qualified_names = False" % dom_section
        snippet_file = "/etc/sssd/conf.d/01_snippet.conf"
        user = (''.join(choice(ascii_uppercase) for _ in range(10)))
        group = (''.join(choice(ascii_uppercase) for _ in range(10)))
        user_group = '{}:{}'.format(user, group)
        multihost.client[0].put_file_contents(snippet_file, file_content)
        cmd_chmod = 'chmod 600 %s' % snippet_file
        multihost.client[0].run_command(cmd_chmod, raiseonerr=False)
        useradd = 'useradd %s' % user
        groupadd = 'groupadd %s' % group
        cmd_chown = 'chown %s %s' % (user_group, snippet_file)
        multihost.client[0].run_command(useradd, raiseonerr=False)
        multihost.client[0].run_command(groupadd, raiseonerr=False)
        multihost.client[0].run_command(cmd_chown, raiseonerr=False)
        start = multihost.client[0].service_sssd('start')
        if start == 0:
            config_check = 'sssctl config-check'
            cmd = multihost.client[0].run_command(config_check,
                                                  raiseonerr=False)
            log_1 = re.compile(r'.*did\snot\spass\saccess\scheck.*')
            if log_1.search(cmd.stdout_text):
                user = 'foo1@%s' % ds_instance_name
                getent1 = 'getent passwd %s' % user
                lookup1 = multihost.client[0].run_command(getent1,
                                                          raiseonerr=False)
                assert lookup1.returncode == 0
                user = 'foo1'
                getent2 = 'getent passwd %s' % user
                lookup2 = multihost.client[0].run_command(getent2,
                                                          raiseonerr=False)
                assert lookup2.returncode == 2
        remove_snippet = 'rm -f %s' % snippet_file
        multihost.client[0].run_command(remove_snippet, raiseonerr=False)
        userdel = 'userdel %s' % user
        groupdel = 'groupdel %s' % group
        multihost.client[0].run_command(userdel, raiseonerr=False)
        multihost.client[0].run_command(groupdel, raiseonerr=False)

    @pytest.mark.tier1
    def test_0006_bz1372258(self, multihost):
        """
        :title: IDM-SSSD-TC: Configuration merging: Add path to files when
         pattern matching fails
        :id: f5620fe5-0576-41b3-acff-f9fceab13206
        """
        multihost.client[0].service_sssd('stop')
        section = "domain/%s" % ds_instance_name
        for idx in ['True', 'False']:
            content = "[%s]\nuse_fully_qualified_names = %s" % (idx, section)
            snippet_file = "/etc/sssd/conf.d/%s_snippet.conf" % idx
            multihost.client[0].put_file_contents(snippet_file, content)
        sssctl_cmd = "sssctl config-check"
        sssctl_check = multihost.client[0].run_command(sssctl_cmd,
                                                       raiseonerr=False)
        result = sssctl_check.stdout_text.strip()
        assert '/etc/sssd/conf.d/False_snippet.conf' in result
        for idx in ['True', 'False']:
            snippet_file = "/etc/sssd/conf.d/%s_snippet.conf" % idx
            cmd = "rm -f %s" % snippet_file
            multihost.client[0].run_command(cmd, raiseonerr=False)

    @pytest.mark.tier1
    def test_0007_bz1466503(self, multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: Configuration merging: Verify scenario when no
         sssd.conf and the snippets should be working
        :id: 0148559d-19d0-4909-b7f2-9d8fc8165db4
        :customerscenario: True
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        # Delete the sssd.conf and add under /etc/sssd/conf.d/ directory
        cp_cmd = 'cp -f /etc/sssd/sssd.conf /etc/sssd/conf.d/01.conf'
        rm_cmd = 'rm -f /etc/sssd/sssd.conf'
        chmod_cmd = 'chmod 600 /etc/sssd/conf.d/01.conf'
        for cmd in [cp_cmd, rm_cmd, chmod_cmd]:
            multihost.client[0].run_command(cmd, raiseonerr=False)
        tools.clear_sssd_cache()
        cmd_start = multihost.client[0].service_sssd('start')
        assert cmd_start == 0
        user = 'foo1@%s' % ds_instance_name
        getent1 = 'getent passwd %s' % user
        lookup1 = multihost.client[0].run_command(getent1,
                                                  raiseonerr=False)
        assert lookup1.returncode == 0
        cmd = "rm -f /etc/sssd/conf.d/01.conf"
        multihost.client[0].run_command(cmd, raiseonerr=False)

    @pytest.mark.tier1
    def test_0008_bz1466503(self, multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: Configuration merging: Verify scenario when no
         sssd.conf and wrong snippets should not be working
        :id: 4b4c340b-0b1d-419d-9347-66e9b068bfff
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        rm_cmd = 'rm -f /etc/sssd/sssd.conf'
        file_content = "[sssd]\ndomains = foo\n\n[domain/foo]\n" \
                       "id_provider = foo"
        snippet_file = "/etc/sssd/conf.d/01_snippet.conf"
        multihost.client[0].put_file_contents(snippet_file, file_content)
        cmd_chmod = 'chmod 600 %s' % snippet_file
        for cmd in [rm_cmd, cmd_chmod]:
            multihost.client[0].run_command(cmd, raiseonerr=False)
        tools.remove_sss_cache('/var/lib/sss/db')
        tools.remove_sss_cache('/var/log/sssd')
        start_sssd = 'systemctl start sssd'
        start = multihost.client[0].run_command(start_sssd, raiseonerr=False)
        assert start.returncode != 0
        cmd_rm = 'rm -f /etc/sssd/conf.d/01_snippet.conf'
        multihost.client[0].run_command(cmd_rm, raiseonerr=False)

    @pytest.mark.tier1
    def test_0009_bz1666307(self, multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: Configuration merging: sssctl config-check
         giving the wrong error message when there are only snippet files
         and no sssd. conf
        :id: e7ca51e4-c41e-41ba-8335-bcc35ec43867
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        # Delete the sssd.conf and add under /etc/sssd/conf.d/ directory
        cp_cmd = 'cp -f /etc/sssd/sssd.conf /etc/sssd/conf.d/01.conf'
        rm_cmd = 'rm -f /etc/sssd/sssd.conf'
        chmod_cmd = 'chmod 600 /etc/sssd/conf.d/01.conf'
        for cmd in [cp_cmd, rm_cmd, chmod_cmd]:
            multihost.client[0].run_command(cmd, raiseonerr=False)
        tools.clear_sssd_cache()
        cmd_start = multihost.client[0].service_sssd('start')
        assert cmd_start == 0
        sssctl_cmd = "sssctl config-check"
        sssctl_check = multihost.client[0].run_command(sssctl_cmd,
                                                       raiseonerr=False)
        result = sssctl_check.stdout_text.strip()
        assert '/etc/sssd/conf.d/01.conf' in result and \
               sssctl_check.returncode == 0
        user = 'foo1@%s' % ds_instance_name
        getent1 = 'getent passwd %s' % user
        lookup1 = multihost.client[0].run_command(getent1,
                                                  raiseonerr=False)
        assert lookup1.returncode == 0
        cmd = "rm -f /etc/sssd/conf.d/01.conf"
        multihost.client[0].run_command(cmd, raiseonerr=False)

    @pytest.mark.tier1
    def test_0010_bz1723273(self, multihost, backupsssdconf):
        """
        :title: IDM-SSSD-TC: Configuration merging: Verify error in snippet
         file created under /etc/sssd/conf.d and /tmp/test/conf.d/
        :id: 5ee4557b-952a-45b3-98f7-7fc98715c53b
        :customerscenario: True
        """
        multihost.client[0].service_sssd('stop')
        cp_cmd = "mkdir -p /tmp/test/conf.d; cp /etc/sssd/sssd.conf /tmp/test/"
        chmod_cmd = "chmod 600 /tmp/test/sssd.conf"
        file_content = "[domain/example1]\n_fully_quailified_name = False\n"
        for snip in ['/etc/sssd/conf.d/', '/tmp/test/conf.d/']:
            for cmd in [cp_cmd, chmod_cmd]:
                multihost.client[0].run_command(cmd, raiseonerr=False)
            snippet_file = snip + '01_snippet.conf'
            multihost.client[0].put_file_contents(snippet_file, file_content)
            cmd_chmod = 'chmod 600 %s' % snippet_file
            multihost.client[0].run_command(cmd_chmod, raiseonerr=False)
            multihost.client[0].service_sssd('start')
            sssctl_cmd = "sssctl config-check -c /tmp/test/sssd.conf -s %s" \
                         % snip
            sssctl_check = multihost.client[0].run_command(sssctl_cmd,
                                                           raiseonerr=False)
            result = sssctl_check.stdout_text.strip()
            cmd_rm = 'rm -f %s' % snippet_file
            multihost.client[0].run_command(cmd_rm, raiseonerr=False)
            rm_dir = 'rm -rf /tmp/test'
            multihost.client[0].run_command(rm_dir, raiseonerr=False)
            assert "Attribute '_fully_quailified_name' is not allowed" \
                   in result and sssctl_check.returncode == 1
